from ..op import Flip,Rotate90,ReduceMean,Expand
from ..interface import Operator, element_wise_shape,numpy
from .compute import \
    ConvolutionMode, \
    compute_convolution_nd, \
    compute_max_pooling_nd, \
    compute_max_unpooling_nd, \
    compute_average_pooling_nd, \
    compute_average_unpooling_nd
from .function import \
    convolution_nd, \
    max_pooling_nd, \
    max_unpooling_nd, \
    average_pooling_nd, \
    average_unpooling_nd


def convolution_nd_shape(shape_data, shape_kernel, dimension, mode):
    prefix_shape = shape_data[:-dimension] + shape_kernel[:-dimension]
    if mode == 'valid' or mode == ConvolutionMode.valid:
        new_shape = prefix_shape + tuple(shape_data[i] - shape_kernel[i] + 1 for i in range(-dimension, 0))
    elif mode == 'full' or mode == ConvolutionMode.full:
        new_shape = prefix_shape + tuple(shape_data[i] + shape_kernel[i] - 1 for i in range(-dimension, 0))
    else:
        raise ValueError('Invalid convolution mode: {}'.format(mode))
    return new_shape, (), ()


def element_wise_convolution_nd_shape(shape_data, shape_kernel, dimension, mode):
    prefix_shape, prefix_broadcast_data, prefix_broadcast_kernel = element_wise_shape(shape_data[:-dimension], shape_kernel[:-dimension])
    if mode == 'valid' or mode == ConvolutionMode.valid:
        new_shape = prefix_shape + tuple(shape_data[i] - shape_kernel[i] + 1 for i in range(-dimension, 0))
    elif mode == 'full' or mode == ConvolutionMode.full:
        new_shape = prefix_shape + tuple(shape_data[i] + shape_kernel[i] - 1 for i in range(-dimension, 0))
    else:
        raise ValueError('Invalid convolution mode: {}'.format(mode))
    return new_shape, prefix_broadcast_data + (0,) * dimension, prefix_broadcast_kernel + (0,) * dimension


def pooling_nd_shape(shape_data, size, step, dimension):
    prefix_shape = shape_data[:-dimension]
    if not isinstance(size, tuple):
        size = (size,)
    if not isinstance(step, tuple):
        step = (step,)
    new_shape = prefix_shape + tuple(len(range(0, shape_data[i] - size[i] + 1, step[i])) for i in range(-dimension, 0))
    return new_shape, ()


def max_unpooling_nd_shape(shape_data, shape_pooling, dimension):
    prefix_shape, prefix_broadcast_data, prefix_broadcast_kernel = element_wise_shape(shape_data[:-dimension], shape_pooling[:-dimension])
    new_shape = prefix_shape + tuple(shape_data[i] for i in range(-dimension, 0))
    return new_shape, prefix_broadcast_data + (0,) * dimension, prefix_broadcast_kernel + (0,) * dimension


def unpooling_nd_shape(shape_pooling, size, step, unpooling_size, dimension):
    prefix_shape, prefix_broadcast_data, prefix_broadcast_kernel = element_wise_shape(shape_pooling[:-dimension], shape_pooling[:-dimension])
    new_shape = prefix_shape + tuple((size[i] + (shape_pooling[i] - 1) * step[i]) if unpooling_size is None else unpooling_size[i] for i in range(-dimension, 0))
    return new_shape, prefix_broadcast_data + (0,) * dimension, prefix_broadcast_kernel + (0,) * dimension


class ConvolutionND(Operator):
    def __init__(self, dimension: int, mode, element_wise: bool=False):
        self.inputs_count = 2
        self.arguments = {'dimension': dimension, 'mode': mode, 'element_wise': element_wise}
        assert dimension > 0

    def f(self, value_data, value_kernel):
        res=compute_convolution_nd(value_data, value_kernel, **self.arguments)
        return res

    def bprop(self, engine, symbol_forward, symbol_data, symbol_kernel):
        forward = engine.bprop(symbol_forward)
        dimension = self.arguments['dimension']
        mode = self.arguments['mode']
        if mode == 'valid' or mode == ConvolutionMode.valid:
            prefix_shape_kernel = engine.shape2(symbol_kernel)[:-dimension]
            prefix_shape_data = engine.shape2(symbol_data)[:-dimension]
            if dimension == 2:
                flip_kernel = symbol_kernel(Rotate90(count=2, axes=(-2, -1)))
            else:
                flip_kernel = symbol_kernel
                for i in range(dimension):
                    flip_kernel = flip_kernel(Flip(-1 - i))
            gradient_data = convolution_nd(forward, flip_kernel, dimension, ConvolutionMode.full, True)
            for _ in prefix_shape_kernel:
                gradient_data = gradient_data(ReduceMean(axis=-dimension - 1))
                symbol_data = symbol_data(Expand(-dimension - 1))
            gradient_kernel = convolution_nd(symbol_data, forward, dimension, ConvolutionMode.valid, True)
            for _ in prefix_shape_data:
                gradient_kernel = gradient_kernel(ReduceMean(axis=-dimension - 1 - len(prefix_shape_kernel)))
            return [lambda: gradient_data,
                    lambda: gradient_kernel]
        elif mode == 'full' or mode == ConvolutionMode.full:
            prefix_shape_kernel = engine.shape2(symbol_kernel)[:-dimension]
            prefix_shape_data = engine.shape2(symbol_data)[:-dimension]
            if dimension == 2:
                flip_kernel = symbol_kernel(Rotate90(count=2, axes=(-2, -1)))
            else:
                flip_kernel = symbol_kernel
                for i in range(dimension):
                    flip_kernel = flip_kernel(Flip(-1 - i))
            gradient_data = convolution_nd(forward, flip_kernel, dimension, ConvolutionMode.valid, True)
            for _ in prefix_shape_kernel:
                gradient_data = gradient_data(ReduceMean(axis=-dimension - 1))
                symbol_data = symbol_data(Expand(-dimension - 1))
            flip_gradient = convolution_nd(forward, symbol_data, dimension, ConvolutionMode.valid, True)
            if dimension == 2:
                gradient_kernel = flip_gradient(Rotate90(count=2, axes=(-2, -1)))
            else:
                gradient_kernel = flip_gradient
                for i in range(dimension):
                    gradient_kernel = gradient_kernel(Flip(-1 - i))
            for _ in prefix_shape_data:
                gradient_kernel = gradient_kernel(ReduceMean(axis=-dimension - 1 - len(prefix_shape_kernel)))
            return [lambda: gradient_data,
                    lambda: gradient_kernel]
        else:
            raise ValueError('Invalid mode: {}'.format(mode))

    def shape(self, shape_data, shape_kernel):
        if self.arguments['element_wise']:
            return element_wise_convolution_nd_shape(shape_data, shape_kernel, self.arguments['dimension'], self.arguments['mode'])
        else:
            return convolution_nd_shape(shape_data, shape_kernel, self.arguments['dimension'], self.arguments['mode'])


class Convolution1D(ConvolutionND):
    def __init__(self, mode, element_wise: bool=False):
        ConvolutionND.__init__(self, 1, mode, element_wise)


class Convolution2D(ConvolutionND):
    def __init__(self, mode, element_wise: bool=False):
        ConvolutionND.__init__(self, 2, mode, element_wise)


class Convolution3D(ConvolutionND):
    def __init__(self, mode, element_wise: bool=False):
        ConvolutionND.__init__(self, 3, mode, element_wise)


class MaxPoolingND(Operator):
    def __init__(self, dimension: int, size: tuple, step: tuple):
        self.inputs_count = 1
        self.arguments = {'dimension': dimension, 'size': size, 'step': step}

    def f(self, value_data):
        res=compute_max_pooling_nd(value_data, **self.arguments)
        return res

    def bprop(self, engine, symbol_forward, symbol_data):
        forward = engine.bprop(symbol_forward)
        return [lambda: max_unpooling_nd(symbol_data, forward, **self.arguments)]

    def shape(self, shape_data):
        return pooling_nd_shape(shape_data, **self.arguments)


class MaxPooling1D(MaxPoolingND):
    def __init__(self, size: tuple, step: tuple):
        MaxPoolingND.__init__(self, 1, size, step)


class MaxPooling2D(MaxPoolingND):
    def __init__(self, size: tuple, step: tuple):
        MaxPoolingND.__init__(self, 2, size, step)


class MaxPooling3D(MaxPoolingND):
    def __init__(self, size: tuple, step: tuple):
        MaxPoolingND.__init__(self, 3, size, step)


class MaxReferencePoolingND(Operator):
    def __init__(self, dimension: int, size: tuple, step: tuple):
        self.inputs_count = 2
        self.arguments = {'dimension': dimension, 'size': size, 'step': step}

    def f(self, value_data, reference_data):
        res=compute_max_pooling_nd(value_data, reference=reference_data, **self.arguments)
        return res

    def bprop(self, engine, symbol_forward, symbol_data, symbol_reference):
        forward = engine.bprop(symbol_forward)
        return [lambda: max_unpooling_nd(symbol_reference, forward, **self.arguments),
                lambda: max_unpooling_nd(symbol_reference, numpy.ones(engine.shape2(forward)), **self.arguments)]

    def shape(self, shape_data):
        return pooling_nd_shape(shape_data, **self.arguments)


class MaxReferencePooling1D(MaxReferencePoolingND):
    def __init__(self, size: tuple, step: tuple):
        MaxReferencePoolingND.__init__(self, 1, size, step)


class MaxReferencePooling2D(MaxReferencePoolingND):
    def __init__(self, size: tuple, step: tuple):
        MaxReferencePoolingND.__init__(self, 2, size, step)


class MaxReferencePooling3D(MaxReferencePoolingND):
    def __init__(self, size: tuple, step: tuple):
        MaxReferencePoolingND.__init__(self, 3, size, step)


class MaxUnpoolingND(Operator):
    def __init__(self, dimension: int, size: tuple, step: tuple):
        self.inputs_count = 2
        self.arguments = {'dimension': dimension, 'size': size, 'step': step}

    def f(self, value_data, value_pooling):
        res=compute_max_unpooling_nd(value_data, value_pooling, **self.arguments)
        return res

    def bprop(self, engine, symbol_forward, symbol_data, symbol_pooling):
        forward = engine.bprop(symbol_forward)
        return [lambda: max_unpooling_nd(symbol_data, numpy.ones(engine.shape2(symbol_pooling)), **self.arguments),
                lambda: max_pooling_nd(forward, reference=symbol_data, **self.arguments)]

    def shape(self, shape_data, shape_pooling):
        return max_unpooling_nd_shape(shape_data, shape_pooling, self.arguments['dimension'])


class MaxUnpooling1D(MaxUnpoolingND):
    def __init__(self, size: tuple, step: tuple):
        MaxUnpoolingND.__init__(self, 1, size, step)


class MaxUnpooling2D(MaxUnpoolingND):
    def __init__(self, size: tuple, step: tuple):
        MaxUnpoolingND.__init__(self, 2, size, step)


class MaxUnpooling3D(MaxUnpoolingND):
    def __init__(self, size: tuple, step: tuple):
        MaxUnpoolingND.__init__(self, 3, size, step)


class AveragePoolingND(Operator):
    def __init__(self, dimension: int, size: tuple, step: tuple):
        self.inputs_count = 1
        self.arguments = {'dimension': dimension, 'size': size, 'step': step}

    def f(self, value_data):
        return compute_average_pooling_nd(value_data, **self.arguments)

    def bprop(self, engine, symbol_forward, symbol_data):
        forward = engine.bprop(symbol_forward)
        return [lambda: average_unpooling_nd(forward, unpooling_size=engine.shape2(symbol_data)[-self.arguments['dimension']:], **self.arguments)]

    def shape(self, shape_data):
        return pooling_nd_shape(shape_data, **self.arguments)


class AveragePooling1D(AveragePoolingND):
    def __init__(self, size: tuple, step: tuple):
        AveragePoolingND.__init__(self, 1, size, step)


class AveragePooling2D(AveragePoolingND):
    def __init__(self, size: tuple, step: tuple):
        AveragePoolingND.__init__(self, 2, size, step)


class AveragePooling3D(AveragePoolingND):
    def __init__(self, size: tuple, step: tuple):
        AveragePoolingND.__init__(self, 3, size, step)


class AverageUnpoolingND(Operator):
    def __init__(self, dimension: int, size: tuple, step: tuple, unpooling_size: tuple=None):
        self.inputs_count = 1
        self.arguments = {'dimension': dimension, 'size': size, 'step': step, 'unpooling_size': unpooling_size}

    def f(self, value_pooling):
        return compute_average_unpooling_nd(value_pooling, **self.arguments)

    def bprop(self, engine, symbol_forward, symbol_pooling):
        forward = engine.bprop(symbol_forward)
        return [lambda: average_pooling_nd(forward, **self.arguments)]

    def shape(self, shape_data):
        return unpooling_nd_shape(shape_data, **self.arguments)


class AverageUnpooling1D(AverageUnpoolingND):
    def __init__(self, size: tuple, step: tuple, unpooling_size: tuple=None):
        AverageUnpoolingND.__init__(self, 1, size, step, unpooling_size)


class AverageUnpooling2D(AverageUnpoolingND):
    def __init__(self, size: tuple, step: tuple, unpooling_size: tuple=None):
        AverageUnpoolingND.__init__(self, 2, size, step, unpooling_size)


class AverageUnpooling3D(AverageUnpoolingND):
    def __init__(self, size: tuple, step: tuple, unpooling_size: tuple=None):
        AverageUnpoolingND.__init__(self, 3, size, step, unpooling_size)



import numpy
def element_wise_shape(*shape_list):
    broadcast_map = {_shape: [] for _shape in shape_list}
    new_shape = []
    for shape in shape_list:#找最大len的shape
        if len(shape) > len(new_shape):
            new_shape = list(shape)
    for i in range(-len(new_shape), 0):
        index = len(new_shape) + i
        dimensions = {}
        for shape in shape_list:
            if -i > len(shape):
                broadcast_map[shape].append(-1)
            else:
                broadcast_map[shape].append(0)
                dimensions[shape] = shape[i]
        new_shape[index] = max([_d for _, _d in dimensions.items()])
        for shape, dimension in dimensions.items():
            if dimension != new_shape[index]:
                if dimension == 1:
                    broadcast_map[shape][-1] = 1
                else:
                    raise ValueError('Can not broadcast these shapes: {}'.format(shape_list))
    return (tuple(new_shape),) + tuple(tuple(broadcast_map[_shape]) for _shape in shape_list)

def matrix_multiply_shape(shape_a, shape_b):
    try:
        if len(shape_a) == 0 or len(shape_b) == 0:
            raise ValueError()
        if len(shape_a) == 1 and len(shape_b) == 1:
            if shape_a[0] == shape_b[0]:
                return (), (), ()
            else:
                raise ValueError()
        if len(shape_a) == 1:
            if shape_a[0] == shape_b[-2]:
                new_shape = list(shape_b)
                del new_shape[-2]
                return tuple(new_shape), (), ()
            else:
                raise ValueError()
        if len(shape_b) == 1:
            if shape_a[-1] == shape_b[0]:
                return shape_a[:-1], (), ()
            else:
                raise ValueError()
        if shape_a[-1] == shape_b[-2]:
            gap = abs(len(shape_a) - len(shape_b))
            if len(shape_a) > len(shape_b):
                if shape_a[gap:-2] != shape_b[:-2]:
                    raise ValueError()
                new_shape = list(shape_a)
                broadcast_a = (0,) * len(shape_a)
                broadcast_b = shape_a[:gap] + (0, 0)
            else:
                if shape_b[gap:-2] != shape_a[:-2]:
                    raise ValueError()
                new_shape = list(shape_b)
                broadcast_a = shape_b[:gap] + (0, 0)
                broadcast_b = (0,) * len(shape_b)
            new_shape[-1] = shape_b[-1]
            new_shape[-2] = shape_a[-2]
            return tuple(new_shape), broadcast_a, broadcast_b
        else:
            raise ValueError()
    except ValueError:
        raise ValueError('Can not execute matrix multiply with these two shapes: a={}, b={}'.format(shape_a, shape_b))


def reduce_shape(shape_a, axis, invariant):
    if axis is None:
        return (), ()
    else:
        new_shape = list(shape_a)
        if invariant:
            new_shape[axis] = 1
        else:
            del new_shape[axis]
        return tuple(new_shape), ()


def transpose_shape(shape_a, axes):
    if axes is None:
        return tuple(reversed(shape_a)), ()
    else:
        if len(set(axes)) == len(axes):
            if set(axes) == set(range(len(axes))) and len(axes) == len(shape_a):
                new_shape = [0] * len(shape_a)
                for i, d in zip(axes, shape_a):
                    new_shape[i] = d
                return tuple(new_shape), ()
            else:
                raise ValueError('Invalid axes for this Shape: shape={}, axes={}'.format(shape_a, axes))
        else:
            raise ValueError('Repeated axis in axes: {}'.format(axes))


def concatenate_shape(axis, *shape_list):
    new_shape = list(shape_list[0])
    shape_len = len(new_shape)
    for shape in shape_list[1:]:
        if len(shape) == len(shape):
            for i in range(shape_len):
                if i == axis:
                    new_shape[i] += shape[i]
                else:
                    if new_shape[i] != shape[i]:
                        raise ValueError('Concatenate shape not match: {}'.format(shape_list))
        else:
            raise ValueError('All shapes must have same dimensions')
    return (tuple(new_shape),) + () * len(shape_list)


def slice_shape(shape_a, slice_list):
    if not isinstance(slice_list, list):
        slice_list = [slice_list]
    new_shape = list(shape_a)
    delete_dimension = 0
    for i, s in enumerate(slice_list):
        index = i - delete_dimension
        if index < len(shape_a):
            if isinstance(s, slice):
                new_shape[index] = len(([0] * shape_a[index])[s])
            elif isinstance(s, int):
                del new_shape[index]
                delete_dimension += 1
            else:
                raise ValueError('Invalid slice type: {}'.format(type(s)))
        else:
            raise ValueError('Shape not match slice: shape={} slice={}'.format(shape_a, slice_list))
    return tuple(new_shape), ()


def rotate90_shape(shape_a, count, axes):
    new_shape = list(shape_a)
    if count % 2 == 0:
        return tuple(new_shape), ()
    else:
        new_shape[axes[0]], new_shape[axes[1]] = new_shape[axes[1]], new_shape[axes[0]]
        return tuple(new_shape), ()

class VCategory:
    variable = 0
    constant = 1
    placeholder = 2
    operator = 3
    
class Operator:
    operator_sign = None
    inputs_count = None
    auto_reduce = True
    arguments = {}

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, ', '.join(['{}={}'.format(key, value) for key, value in self.arguments.items()]))

    def f(self, *args, **kwargs):
        raise ValueError("this is abstract")

    def bprop(self, *args, **kwargs):
        raise ValueError("this is abstract")

    def shape(self, *args, **kwargs):
        raise ValueError("this is abstract")

class Template:
    active_operator = None
    def __init__(self):
        if self.active_operator is None:
            self.active_sign = None
        else:
            self.active_sign = self.active_operator.__name__

    @staticmethod
    def reduce_symbol(symbol, index: int):
        from .tensor import V
        if not isinstance(symbol,V):raise ValueError("symbol Must be V")
        input_list = symbol.input
        reduce_to_symbol = input_list[index]
        symbol.clear_operator()
        symbol.value = reduce_to_symbol.value
        symbol.name = reduce_to_symbol.name
        if reduce_to_symbol.is_operator():
            symbol.symbolic_compute(reduce_to_symbol.operator, reduce_to_symbol.input)
        else:
            symbol.category = reduce_to_symbol.category

    @staticmethod
    def value_equal(a, b):
        result = a == b
        if isinstance(result, bool) or isinstance(result, numpy.bool_):
            return result
        elif isinstance(result, numpy.ndarray):
            return result.all()
        else:
            raise Exception('Never reached.')

    @staticmethod
    def symbol_equal(a, b):
        from .tensor import V
        if not isinstance(a,V) or not isinstance(b,V):raise ValueError("symbol Must be V")
        return a.symbolic_hash() == b.symbolic_hash()

    def simplify(self, symbol):
        raise ValueError("this is abstract")

class Simplification:
    def __init__(self,templates:list):
        self.__templates = {}
        for template in templates:
            self.register(template())
    def operator_trigger(self, operator: Operator):
        operator_sign = operator.__class__.__name__
        if operator_sign in self.__templates:
            return self.__templates[operator_sign]
        else:
            return set()

    def register(self, template: Template):
        active_operator = template.active_sign
        self.__templates.setdefault(active_operator, set())
        self.__templates[active_operator].add(template)

    def simplify(self, symbol):
        from .tensor import V
        if not isinstance(symbol,V):raise ValueError("symbol Must be V")
        while self.simplify_cycle(symbol):
            pass
    def simplify_cycle(self, symbol):
        from .tensor import V
        if not isinstance(symbol,V):raise ValueError("symbol Must be V")
        effective = False
        templates = list(self.operator_trigger(symbol.operator)) + list(self.__templates[None])
        if templates:
            for template in templates:
                if template.simplify(symbol):
                    effective |= True
                    break
        for next_symbol in symbol.input:
            if next_symbol.operator is not None:
                effective |= self.simplify_cycle(next_symbol)
        return effective
    
class Optimizer:
    def optimize(self, engine,calculate_function):
        raise ValueError("this is abstract")
    def minimize(self, engine,steps: int=1):
        for _ in range(steps):
            self.optimize(engine,lambda v, g: v - g)


from .interface import Operator,VCategory,numpy
class V:
    def __init__(self, value=None, shape: tuple=None, name: str=None, operator=None, inputs=None, category: VCategory=None):
        self.__name = None
        self.__input = []
        self.__operator = None
        self.__output = []
        self.__value = None
        self.__shape = None
        self.__scalar = False
        self.__category = None
        # for engin
        self.__variables = set()
        self.__grad_table = {}
        self.__shape2 = {}
        self.__broadcast = {}
        self.__bind ={}
        self.__value_cache = {}
        if isinstance(value, V):
            self.__set_name(value.name)
            self.__set_value(value.value)
            self.__set_shape(value.shape)
            for _input in value.input:
                self.__add_input(_input.clone())
            self.__set_operator(value.__operator)
            for _output in value.output:
                self.__add_output(_output.clone())
            self.__set_category(value.category)
        else:
            self.symbolic_compute(operator, inputs)
            self.__set_value(value)
            self.__set_category(category)
            self.__set_shape(shape)
            self.__set_name(name)
    def __call__(self,op,*arg):
        arg=[self,]+list(arg)
        res=V(operator=op,inputs=as_symbols(arg))
        simp.simplify(res)
        return res
    def __repr__(self):
        if self.__operator is None:
            return self.__name
        else:
            if self.__operator.operator_sign is None:
                arguments = list(map(str, self.input))
                arguments += ['{}={}'.format(k, '\'' + v + '\'' if isinstance(v, str) else v) for k, v in self.__operator.arguments.items()]
                return '{}({})'.format(self.__operator.__class__.__name__, ', '.join(arguments))
            else:
                if len(self.input) < 2:
                    return '{}{}'.format(self.__operator.operator_sign, self.input[0])
                else:
                    return '({} {} {})'.format(self.input[0], self.__operator.operator_sign, self.input[1])

    def __get_name(self):
        return self.__name

    def __set_name(self, name: str):
        if name is None:
            if self.__value is None:
                self.__name = self.__class__.__name__
            else:
                if self.is_scalar():
                    self.__name = str(self.__value)
                else:
                    self.__name = '{}{}'.format(self.__class__.__name__, self.__value.shape)
        else:
            self.__name = name

    name = property(__get_name, __set_name)

    def __get_category(self):
        return self.__category

    def __set_category(self, category: VCategory):
        if category is None:
            if self.__category is None:
                self.__category = VCategory.variable
        else:
            if self.__category == VCategory.operator:
                if category != VCategory.operator:
                    raise ValueError('Can not change category of Operator V.')
            if category == VCategory.variable:
                self.__category = category
            elif category == VCategory.constant:
                self.__category = category
                if self.__value is None:
                    raise ValueError('constant V must have value.')
            elif category == VCategory.placeholder:
                self.__category = category
                if self.__value is not None:
                    self.__set_shape(self.__value.shape)
            elif category == VCategory.operator:
                if self.operator is not None:
                    self.__category = category
                else:
                    raise ValueError('Can not convert other V to Operator V.')
            else:
                raise ValueError('Invalid category: {}'.format(category))

    category = property(__get_category, __set_category)

    def __get_value(self):
        return self.__value

    def __set_value(self, tensor):
        if tensor is not None:
            if self.__category == VCategory.constant:
                raise ValueError('Can not change value for constant.')
            else:
                if self.__operator is None:
                    self.__value = numpy.array(tensor, dtype=float)
                    self.__scalar = len(self.value.shape) == 0
                    self.__set_category(VCategory.variable)
                else:
                    raise ValueError('Can not assign value for Operator V.')

    value = property(__get_value, __set_value)

    def __get_shape(self):
        if self.__category == VCategory.placeholder:
            return self.__shape
        else:
            if self.__operator is None:
                return self.__value.shape
            else:
                return self.shape2(self)
#                 raise ValueError('Operator V has no shape.')

    def __set_shape(self, shape: tuple):
        if shape is not None:
            if self.__category == VCategory.placeholder:
                self.__shape = shape
            else:
                raise ValueError('Only Placeholder can set shape.')

    shape = property(__get_shape, __set_shape)

    def __get_operator(self):
        return self.__operator

    def __set_operator(self, operator):
        if operator is not None:
            if isinstance(operator, Operator):
                self.__operator = operator
            else:
                raise ValueError('Operator must be Operator.')

    operator = property(__get_operator, __set_operator)

    def __get_input(self):
        return self.__input

    input = property(__get_input)

    def __add_input(self, symbol):
        if isinstance(symbol, V):
            self.__input.append(symbol)
        else:
            raise ValueError('Input must be V.')

    def __get_output(self):
        return self.__output

    output = property(__get_output)

    def __add_output(self, symbol):
        if isinstance(symbol, V):
            self.__output.append(symbol)
        else:
            raise ValueError('Output must be V.')

    def symbolic_compute(self, operator, inputs):
        if operator is not None and inputs:
            self.__set_operator(operator)
            inputs_count = operator.inputs_count
            if inputs_count is None:
                inputs_count = len(inputs)
            self.__input = []
            self.__scalar = True
            for symbol in inputs[:inputs_count]:
                if isinstance(symbol, V):
                    if not symbol.is_scalar():
                        self.__scalar = False
                    self.__add_input(symbol)
                    symbol.__add_output(self)
                else:
                    raise ValueError('Input must be V.')
            self.__set_category(VCategory.operator)

    def remove_input(self, symbol):
        new_input = []
        find_input = None
        for each_input in self.__input:
            if hash(each_input) != hash(symbol):
                new_input.append(each_input)
            else:
                find_input = each_input
        self.__input = new_input
        if find_input is not None:
            find_input.remove_output(self)

    def remove_output(self, symbol):
        new_output = []
        find_output = None
        for each_output in self.__output:
            if hash(each_output) != hash(symbol):
                new_output.append(each_output)
            else:
                find_output = each_output
        self.__output = new_output
        if find_output is not None:
            find_output.remove_input(self)

    def clear_input(self):
        for symbol in set(self.__input):
            symbol.remove_output(self)
        self.__input = []

    def clear_output(self):
        for symbol in set(self.__output):
            symbol.remove_input(self)
        self.__output = []

    def clear_operator(self):
        self.clear_input()
        self.__operator = None
        self.__category = VCategory.variable

    def destroy(self):
        self.clear_input()
        self.clear_output()
        self.__value = None
        self.__operator = None

    def is_scalar(self):
        return self.__scalar
    def is_constant(self):
        return self.__category == VCategory.constant
    def is_variable(self):
        return self.__category == VCategory.variable
    def is_placeholder(self):
        return self.__category == VCategory.placeholder
    def is_operator(self):
        return self.__category == VCategory.operator

    def symbolic_hash(self):
        if self.is_operator():
            inputs_symbolic_hash = [each_input.symbolic_hash() for each_input in self.input]
            return '{}({})'.format(self.operator.__class__.__name__, ','.join(inputs_symbolic_hash))
        else:
            return str(hash(self))

    def __hash__(self):
        return id(self)
    def __getitem__(self, item):
        return slice_select(self, item)
    def __setitem__(self, key, value):
        return slice_assign(self, value, key)
    
    def clear(self):
        self.__grad_table = {}
        self.__shape2 = {}
        self.__broadcast = {}
        self.__value_cache = {}
    def get_variables(self):
        return self.__variables

    def set_variables(self, symbol):
        if symbol is None:
            symbol = set()
            symbol_set ={self} 
            while len(symbol_set):
                any_symbol = symbol_set.pop()
                if any_symbol.is_variable():
                    symbol.add(any_symbol)
                elif any_symbol.is_operator():
                    symbol_set |= set(any_symbol.input)
        old_variables = set(self.__variables)
        if isinstance(symbol, V):
            symbols = {symbol}
        else:
            symbols = set(symbol)
        for symbol in symbols:
            if isinstance(symbol, V) and not symbol.is_operator():
                self.__variables.add(symbol)
            else:
                raise ValueError('Variable must be V.')
        unused_variables = old_variables - self.__variables
        for variable in unused_variables:
            if variable in self.__grad_table:
                del self.__grad_table[variable]

    variables = property(get_variables, set_variables)

    def get_bind(self,):
        return self.__bind

    def set_bind(self,bind_data: dict):
        old_bind = self.__bind
        self.__bind ={}
        need_clear = False
        for s, d in bind_data.items():
            if s.category == VCategory.constant:
                raise ValueError('Can not bind data for constant.')
            d_array = numpy.array(d)
            if s in old_bind:
                if old_bind[s].shape != d_array.shape:
                    need_clear = True
            else:
                need_clear = True
            self.__bind[s] = d_array
        if need_clear:
            self.clear()
    bind=property(get_bind,set_bind)
    def get_value_cache(self):
        return self.__value_cache

    def set_value_cache(self, value_cache: dict):
        self.__value_cache = value_cache

    value_cache = property(get_value_cache, set_value_cache)

    def modified(self):
        self.__value_cache = {}

    def __compute_value(self,symbol):
        if not isinstance(symbol,V):raise ValueError('symbol Must be V')
        if not symbol.is_operator():
            if symbol in self.__bind:
                return numpy.array(self.__bind[symbol])
            else:
                if symbol.value is None or symbol.is_placeholder():
                    raise ValueError('V must bind data: {}'.format(symbol))
                else:
                    return symbol.value
        else:
            if symbol in self.__value_cache:
                return self.__value_cache[symbol]
            else:
                compute_inputs =[self.__compute_value(_s) for _s in symbol.input]
                symbol_value = symbol.operator.f(*compute_inputs)
                self.__value_cache[symbol] =symbol_value
                return symbol_value

    def __build_grad(self, variable):
        if not isinstance(variable,V):
            raise ValueError('variable must be V')
        if hash(self) == hash(variable):
            from trik.sheng.op import Broadcast
            self.__grad_table[variable] = constant(1)(Broadcast(self.shape2(self)))
            return
        current_operator = None
        index = -1
        for forward in variable.output:
            if forward.is_constant():continue
            if self.bprop(forward) is not None:
                if current_operator != forward.operator:
                    current_operator = forward.operator
                    index = -1
                gradients = forward.operator.bprop(self, forward, *forward.input)
                for i, _variable in enumerate(forward.input, start=index + 1):
                    if hash(_variable) == hash(variable):
                        index = i
                        break
                if gradients:current_gradient=gradients[index]()
                else:current_gradient=constant(0)
                if forward.operator.auto_reduce:
                    invariant = 0
                    from trik.sheng.op import ReduceSum
                    for i, d in enumerate(self.broadcast(variable, forward)):
                        if d > 0:
                            current_gradient = current_gradient(ReduceSum(axis=i + invariant, invariant=True))
                        elif d < 0:
                            current_gradient = current_gradient(ReduceSum(axis=i + invariant, invariant=False))
                            invariant -= 1
                if variable not in self.__grad_table:
                    self.__grad_table[variable] = current_gradient
                else:
                    from trik.sheng.op import Plus
                    self.__grad_table[variable] = self.__grad_table[variable](Plus(),current_gradient)
        if variable in self.__grad_table:
            simp.simplify(self.__grad_table[variable])

    def __compute_shape(self, symbol):
        if not isinstance(symbol,V):
            raise ValueError('symbol must be V')
        if not symbol.is_operator():
            if symbol in self.__bind:
                self.__shape2[symbol] = self.__bind[symbol].shape
            else:
                if symbol.shape is None:
                    raise ValueError('Placeholder must bind data or set shape: {}'.format(symbol))
                else:
                    self.__shape2[symbol] = symbol.shape
        else:
            shape_broadcasts = symbol.operator.shape(*[self.shape2(s) for s in symbol.input])
            shape = shape_broadcasts[0]
            broadcasts = shape_broadcasts[1:]
            self.__shape2[symbol] = shape
            for input_symbol, input_broadcast in zip(symbol.input, broadcasts):
                if sum([abs(d) for d in input_broadcast]) > 0:
                    self.__broadcast.setdefault(input_symbol, {})
                    self.__broadcast[input_symbol].setdefault(symbol, {})
                    self.__broadcast[input_symbol][symbol] = input_broadcast

    def val(self):
        return self.__compute_value(self)

    def differentiate(self):
        for variable in self.__variables:
            if variable not in self.__grad_table:
                self.__build_grad(variable)

    def bprop(self, variable):
        if not isinstance(variable,V):
            raise ValueError('variable must be V')
        if variable not in self.__grad_table:
            self.__build_grad(variable)
        return self.__grad_table.get(variable, None)

    def shape2(self, variable):
        if not isinstance(variable,V):
            raise ValueError('variable must be V')
        if variable not in self.__shape2:
            self.__compute_shape(variable)
        return self.__shape2.get(variable, None)

    def broadcast(self, from_variable, to_variable):
        if not isinstance(from_variable,V) or not isinstance(to_variable,V):
            raise ValueError('variable must be V')
        if from_variable not in self.__broadcast:
            self.__compute_shape(from_variable)
        if from_variable not in self.__broadcast:
            return ()
        else:
            if to_variable not in self.__broadcast[from_variable]:
                return ()
            else:
                return self.__broadcast[from_variable][to_variable]
    
def constant(value=None, shape: tuple=None, name: str=None, operator=None, inputs=None):
    return V(value, shape, name, operator, inputs, VCategory.constant)
def variable(value=None, shape: tuple=None, name: str=None, operator=None, inputs=None):
    return V(value, shape, name, operator, inputs, VCategory.variable)
def placeholder(value=None, shape: tuple=None, name: str=None, operator=None, inputs=None):
    return V(value, shape, name, operator, inputs, VCategory.placeholder)

def as_symbol(thing):
    if isinstance(thing, V):
        return thing
    else:
        return constant(thing)
def as_symbols(things):
    return list(map(as_symbol, things))
def slice_assign(a, b, slice_tuple):
    from trik.sheng.op import SliceAssign
    return V(operator=SliceAssign(slice_tuple), inputs=as_symbols([a, b]))  
def slice_select(a, slice_tuple):
    from trik.sheng.op import SliceSelect
    return V(operator=SliceSelect(slice_tuple), inputs=as_symbols([a]))
from .ab import simp


import numpy
def array_index_traversal(array_shape):
    scale = numpy.prod(array_shape)
    for i in range(scale):
        index = [0] * len(array_shape)
        index[-1] = i
        for r in range(len(array_shape) - 1, -1, -1):
            if index[r] >= array_shape[r]:
                index[r - 1] = index[r] // array_shape[r]
                index[r] %= array_shape[r]
            else:
                break
        yield tuple(index)


def multi_range(range_list):
    len_list = [len(r) for r in range_list]
    scale = numpy.prod(len_list)
    for i in range(scale):
        index = [0] * len(len_list)
        index[-1] = i
        for r in range(len(len_list) - 1, -1, -1):
            if index[r] >= len_list[r]:
                index[r - 1] = index[r] // len_list[r]
                index[r] %= len_list[r]
            else:
                break
        yield tuple(range_list[_i][_index] for _i, _index in enumerate(index))



from .interface import Template,VCategory,Simplification
from .tensor import V,constant
from .op import MM,SliceSelect
class TemplateConstant(Template):
    active_operator = None

    def simplify(self, symbol: V):
        if symbol.is_operator():
            for s in symbol.input:
                if not s.is_constant():
                    return False
            compute_inputs = [_s.value for _s in symbol.input]
            value = symbol.operator.f(*compute_inputs)
            symbol.clear_operator()
            symbol.value = value
            symbol.category = VCategory.constant
            symbol.name=None
            return True
        else:
            return False


class TemplatePlus(Template):
    from trik.sheng.op import Plus
    active_operator = Plus

    def simplify(self, symbol: V):
        from trik.sheng.op import Mul,Broadcast
        left_symbol, right_symbol = symbol.input
        if left_symbol.is_constant() and self.value_equal(left_symbol.value, 0):
            self.reduce_symbol(symbol, 1)
            return True
        elif right_symbol.is_constant() and self.value_equal(right_symbol.value, 0):
            self.reduce_symbol(symbol, 0)
            return True
        elif self.symbol_equal(left_symbol, right_symbol):
            symbol.clear_operator()
            symbol.symbolic_compute(Mul(), [left_symbol, constant(2)])
            symbol.name=None
            return True
        elif left_symbol.operator!=None: 
            if left_symbol.operator.operator_sign=='-' and left_symbol.operator.inputs_count==1 and self.symbol_equal(left_symbol.input[0],right_symbol):
                symbol.clear_operator()
                symbol.symbolic_compute(Broadcast(()),[constant(0)])
                symbol.name=None
                return True
        elif right_symbol.operator!=None: 
            if right_symbol.operator.operator_sign=='-' and left_symbol.operator.inputs_count==1 and self.symbol_equal(right_symbol.input[0],left_symbol):
                symbol.clear_operator()
                symbol.symbolic_compute(Broadcast(()),[constant(0)])
                symbol.name=None
                return True
        else:
            return False


class TemplateSubtract(Template):
    from trik.sheng.op import Sub
    active_operator = Sub

    def simplify(self, symbol: V):
        from trik.sheng.op import Neg,Broadcast#SliceAssign
        left_symbol, right_symbol = symbol.input
        if right_symbol.is_constant() and self.value_equal(left_symbol.value, 0):
            symbol.clear_operator()
            symbol.clear_input()
            symbol.symbolic_compute(Neg(), [right_symbol])
            symbol.name=None
            return True
        elif right_symbol.is_constant() and self.value_equal(right_symbol.value, 0):
            self.reduce_symbol(symbol, 0)
            return True
        elif self.symbol_equal(left_symbol, right_symbol):
            symbol.clear_operator()
            symbol.symbolic_compute(Broadcast(()), [left_symbol, constant(0)])
            symbol.name=None
            return True
        else:
            return False


class TemplateDivide(Template):
    from trik.sheng.op import Div
    active_operator = Div

    def simplify(self, symbol: V):
        from trik.sheng.op import SliceAssign
        left_symbol, right_symbol = symbol.input
        if right_symbol.is_constant() and self.value_equal(right_symbol.value, 1):
            self.reduce_symbol(symbol, 0)
            return True
        elif left_symbol.is_constant() and self.value_equal(left_symbol.value, 0):
            symbol.clear_operator()
            symbol.symbolic_compute(SliceAssign(slice(None)), [left_symbol, constant(0)])
            symbol.name=None
            return True
        elif self.symbol_equal(left_symbol, right_symbol):
            symbol.clear_operator()
            symbol.symbolic_compute(SliceAssign(slice(None)), [left_symbol, constant(1)])
            symbol.name=None
            return True
        else:
            return False


class TemplateMultiply(Template):
    from trik.sheng.op import Mul
    active_operator = Mul

    def simplify(self, symbol: V):
        from trik.sheng.op import Pow,Broadcast
        left_symbol, right_symbol = symbol.input
        if left_symbol.is_constant() and self.value_equal(left_symbol.value, 1):
            self.reduce_symbol(symbol, 1)
            return True
        elif right_symbol.is_constant() and self.value_equal(right_symbol.value, 1):
            self.reduce_symbol(symbol, 0)
            return True
        elif (left_symbol.is_constant() and self.value_equal(left_symbol.value, 0)) or(right_symbol.is_constant() and self.value_equal(right_symbol.value, 0)) :
            symbol.clear_operator()
            symbol.symbolic_compute(Broadcast(()),[constant(0)])
            symbol.name=None
            return True
        elif self.symbol_equal(left_symbol, right_symbol):
            symbol.clear_operator()
            symbol.symbolic_compute(Pow(), [left_symbol, constant(2)])
            symbol.name=None
            return True
        elif left_symbol.operator!=None: 
            if left_symbol.operator.operator_sign=='/' and left_symbol.input[0].is_constant() and self.symbol_equal(left_symbol.input[1],right_symbol):
                if left_symbol.input[0].value==1:
                    symbol.clear_operator()
                    symbol.symbolic_compute(Broadcast(()),[constant(1)])
                    symbol.name=None
                    return True
        elif right_symbol.operator!=None: 
            if right_symbol.operator.operator_sign=='/' and right_symbol.input[0].is_constant() and self.symbol_equal(right_symbol.input[1],left_symbol):
                if right_symbol.input[0].value==1:
                    symbol.clear_operator()
                    symbol.symbolic_compute(Broadcast(()),[constant(1)])
                    symbol.name=None
                    return True
        else:
            return False


class TemplatePower(Template):
    from trik.sheng.op import Pow
    active_operator = Pow
    def simplify(self, symbol: V):
        from trik.sheng.op import Broadcast
        left_symbol, right_symbol = symbol.input
        if (self.value_equal(left_symbol.value, 1) and left_symbol.is_constant()) or (self.value_equal(right_symbol.value, 1) and right_symbol.is_constant()):
            self.reduce_symbol(symbol, 0)
            return True
        elif self.value_equal(right_symbol.value, 0) and right_symbol.is_constant():
            symbol.clear_operator()
            symbol.symbolic_compute(Broadcast(()),[constant(1)])    
            symbol.name=None
            return True
        else:
            return False

class TemplateMM(Template):#待优化
    active_operator = MM
    maxlen=4;
    def simplify(self, symbol:V):
        ls=g(symbol)
        if len(ls)>self.maxlen:
            self.maxlen=len(ls)
#             print(ls)
            symbol2=rebuildChain(ls)            
            symbol.clear_operator()
            symbol.symbolic_compute(SliceSelect(slice(None,None,None)),[symbol2])    
            symbol.name=None
            return False


def chainTrace(i,j,s):
    if i==j:
        return []
    ls=[]
    ls+=(chainTrace(i,s[(i,j)],s))
    ls+=(chainTrace(s[(i,j)]+1,j,s))
    ls.append(s[(i,j)])
    return ls    
def matChain(p:list):
    n=len(p)-1
    m,s={},{}
    for i in range(1,n+1):m[(i,i)]=0
    for r in range(2,n+1):
        for i in range(1,n-r+2):
            j=i+r-1
            m[(i,j)]=m[(i+1,j)]+p[i-1]*p[i]*p[j]
            s[(i,j)]=i;
            for k in range(i+1,j):
                t=m[(i,k)]+m[(k+1,j)]+p[i-1]*p[k]*p[j]
                if t<m[(i,j)]:
                    m[(i,j)]=t;s[(i,j)]=k    
    return chainTrace(1, n, s)
def rebuildChain(ls):
    if len(ls)<3:return
    ls1=[ls[0].shape[0]]+[i.shape[1] for i in ls]
    m=matChain(ls1)
    nls=ls[m[0]-1](MM(),ls[m[0]])
    ls[m[0]-1]=nls
    ls[m[0]]=None
    for i in m[1:]:
        j=0
        k=1
        while ls[i+j]==None:j+=1
        while ls[i-k]==None:k+=1
        nls=ls[i-k](MM(),ls[i+j])
        ls[i-k]=nls
        ls[i+j]=None
    return nls

def g(l):#待优化
    if l.operator==None or l.operator.operator_sign!='@':
        return [l]
    ls =[]
    for i in range(l.operator.inputs_count):
        ls+=g(l.input[i])
    return ls

default_templates = [
    TemplateConstant,
    TemplatePlus,
    TemplateSubtract,
    TemplateMultiply,
    TemplateDivide,
    TemplatePower,
    TemplateMM,
]
simp=Simplification(default_templates)


from ..tensor import V,as_symbols

def convolution_nd(data, kernel, dimension: int, mode, element_wise: bool=False):
    from .operator import ConvolutionND
    return V(operator=ConvolutionND(dimension, mode, element_wise), inputs=as_symbols([data, kernel]))


def convolution_1d(data, kernel, mode, element_wise: bool=False):
    return convolution_nd(data, kernel, 1, mode, element_wise)


def convolution_2d(data, kernel, mode, element_wise: bool=False):
    return convolution_nd(data, kernel, 2, mode, element_wise)


def convolution_3d(data, kernel, mode, element_wise: bool=False):
    return convolution_nd(data, kernel, 3, mode, element_wise)


def max_pooling_nd(data, size: tuple, step: tuple, dimension: int, reference=None):
    if reference is None:
        from .operator import MaxPoolingND
        return V(operator=MaxPoolingND(dimension, size, step), inputs=as_symbols([data]))
    else:
        from .operator import MaxReferencePoolingND
        return V(operator=MaxReferencePoolingND(dimension, size, step), inputs=as_symbols([data, reference]))


def max_pooling_1d(data, size: tuple, step: tuple, reference=None):
    return max_pooling_nd(data, size, step, 1, reference)


def max_pooling_2d(data, size: tuple, step: tuple, reference=None):
    return max_pooling_nd(data, size, step, 2, reference)


def max_pooling_3d(data, size: tuple, step: tuple, reference=None):
    return max_pooling_nd(data, size, step, 3, reference)


def max_unpooling_nd(data, pooling, size: tuple, step: tuple, dimension: int):
    from .operator import MaxUnpoolingND
    return V(operator=MaxUnpoolingND(dimension, size, step), inputs=as_symbols([data, pooling]))


def max_unpooling_1d(data, pooling, size: tuple, step: tuple):
    return max_unpooling_nd(data, pooling, size, step, 1)


def max_unpooling_2d(data, pooling, size: tuple, step: tuple):
    return max_unpooling_nd(data, pooling, size, step, 2)


def max_unpooling_3d(data, pooling, size: tuple, step: tuple):
    return max_unpooling_nd(data, pooling, size, step, 3)


def average_pooling_nd(data, size: tuple, step: tuple, dimension: int):
    from .operator import AveragePoolingND
    return V(operator=AveragePoolingND(dimension, size, step), inputs=as_symbols([data]))


def average_pooling_1d(data, size: tuple, step: tuple):
    return average_pooling_nd(data, size, step, 1)


def average_pooling_2d(data, size: tuple, step: tuple):
    return average_pooling_nd(data, size, step, 2)


def average_pooling_3d(data, size: tuple, step: tuple):
    return average_pooling_nd(data, size, step, 3)


def average_unpooling_nd(pooling, size: tuple, step: tuple, dimension: int, unpooling_size: int=None):
    from .operator import AverageUnpoolingND
    return V(operator=AverageUnpoolingND(dimension, size, step, unpooling_size), inputs=as_symbols([pooling]))


def average_unpooling_1d(pooling, size: tuple, step: tuple, unpooling_size: tuple=None):
    return average_unpooling_nd(pooling, size, step, 1, unpooling_size)


def average_unpooling_2d(pooling, size: tuple, step: tuple, unpooling_size: tuple=None):
    return average_unpooling_nd(pooling, size, step, 2, unpooling_size)


def average_unpooling_3d(pooling, size: tuple, step: tuple, unpooling_size: tuple=None):
    return average_unpooling_nd(pooling, size, step, 3, unpooling_size)



from functools import reduce
from .tensor import as_symbols,V,constant
from .interface import Operator,numpy, \
    element_wise_shape,matrix_multiply_shape,reduce_shape,transpose_shape,concatenate_shape, \
    slice_shape, rotate90_shape

class Neg(Operator):
    def __init__(self):
        self.operator_sign = '-'
        self.inputs_count = 1

    def f(self, value_a):
        return -value_a

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Neg())]

    def shape(self, shape_a):
        return shape_a, ()


class Abs(Operator):
    def __init__(self):
        self.inputs_count = 1

    def f(self, value_a):
        return numpy.absolute(value_a)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),V(operator=Where(),inputs=as_symbols([symbol_a(Gt(),0), 1, -1])))]

    def shape(self, shape_a):
        return shape_a, ()


class Plus(Operator):
    def __init__(self):
        self.operator_sign = '+'
        self.inputs_count = 2

    def f(self, value_a, value_b):
        return value_a + value_b

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),1),
                lambda: forward(Mul(),1)]

    def shape(self, shape_a, shape_b):
        return element_wise_shape(shape_a, shape_b)


class Sub(Operator):
    def __init__(self):
        self.operator_sign = '-'
        self.inputs_count = 2

    def f(self, value_a, value_b):
        return value_a - value_b

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),1),
                lambda: forward(Mul(),-1)]

    def shape(self, shape_a, shape_b):
        return element_wise_shape(shape_a, shape_b)


class Mul(Operator):
    def __init__(self):
        self.operator_sign = '*'
        self.inputs_count = 2

    def f(self, value_a, value_b):
        return value_a * value_b

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),symbol_b),
                lambda: forward(Mul(),symbol_a)]

    def shape(self, shape_a, shape_b):
        return element_wise_shape(shape_a, shape_b)


class Div(Operator):
    def __init__(self):
        self.operator_sign = '/'
        self.inputs_count = 2

    def f(self, value_a, value_b):
        return value_a / value_b

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),1)(Div(),symbol_b),
                lambda: forward(Mul(),-1)(Mul(),symbol_a)(Div(),symbol_b(Pow(),2))]

    def shape(self, shape_a, shape_b):
        return element_wise_shape(shape_a, shape_b)


class MM(Operator):
    def __init__(self):
        self.operator_sign = '@'
        self.inputs_count = 2

    def f(self, value_a, value_b):
        if len(value_a.shape) == 0 or len(value_b.shape) == 0:
            return value_a * value_b
        else:
            return value_a @ value_b

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        shape_a = engine.shape2(symbol_a)
        shape_b = engine.shape2(symbol_b)
        if len(shape_a) >= 2:
            axes_a = tuple(range(len(shape_a) - 2)) + (-1, -2)
        else:
            axes_a = None
        if len(shape_b) >= 2:
            axes_b = tuple(range(len(shape_b) - 2)) + (-1, -2)
        else:
            axes_b = None
        return [lambda: forward(MM(),symbol_b(T(axes=axes_a))),
                lambda: symbol_a(T(axes=axes_b))(MM(),forward)]

    def shape(self, shape_a, shape_b):
        return matrix_multiply_shape(shape_a, shape_b)


class T(Operator):
    def __init__(self, axes=None):
        self.inputs_count = 1
        self.arguments = {'axes': axes}

    def f(self, value_a):
        return numpy.transpose(value_a, axes=self.arguments['axes'])

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(T(axes=self.arguments['axes']))]

    def shape(self, shape_a):
        return transpose_shape(shape_a, self.arguments['axes'])


class ReduceSum(Operator):
    def __init__(self, axis: int=None, invariant: bool=False):
        self.inputs_count = 1
        self.arguments = {'axis': axis, 'invariant': invariant}

    def f(self, value_a):
        return numpy.sum(value_a, axis=self.arguments['axis'], keepdims=self.arguments['invariant'])

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        shape_a = engine.shape2(symbol_a)
        axis = self.arguments['axis']
        if axis:
            return [lambda: forward(Expand(axis))(Broadcast(shape_a))]
        else:
            return [lambda: forward(Broadcast(shape_a))]

    def shape(self, shape_a):
        return reduce_shape(shape_a, **self.arguments)


class ReduceMean(Operator):
    def __init__(self, axis: int=None, invariant: bool=False):
        self.inputs_count = 1
        self.arguments = {'axis': axis, 'invariant': invariant}

    def f(self, value_a):
        return numpy.mean(value_a, axis=self.arguments['axis'], keepdims=self.arguments['invariant'])

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        shape_a = engine.shape2(symbol_a)
        axis = self.arguments['axis']
        if axis:
            return [lambda: forward(Expand(axis))(Broadcast(shape_a))(Div(),shape_a[axis])]
        else:
            return [lambda: forward(Broadcast(shape_a))( Div(),reduce(lambda x, y: x * y, shape_a, 1))]

    def shape(self, shape_a):
        return reduce_shape(shape_a, **self.arguments)


class Expand(Operator):
    def __init__(self, axis: int):
        self.inputs_count = 1
        self.arguments = {'axis': axis}

    def f(self, value_a):
        return numpy.expand_dims(value_a, **self.arguments)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(ReduceMean(**self.arguments))]

    def shape(self, shape_a):
        new_shape = list(shape_a)
        new_shape.insert(self.arguments['axis'], 1)
        broadcast_a = [0] * len(shape_a)
        broadcast_a.insert(self.arguments['axis'], 1)
        return tuple(new_shape), tuple(broadcast_a)


class Broadcast(Operator):
    def __init__(self, shape):
        self.inputs_count = 1
        self.arguments = {'shape': shape}

    def f(self, value_a):
        if len(value_a.shape) == 1 and len(self.arguments['shape']) > 1:
            if value_a.shape[0] == self.arguments['shape'][0]:
                value_a = value_a.reshape((value_a.shape[0], 1))
        return numpy.broadcast_to(value_a, **self.arguments)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward]

    def shape(self, shape_a):
        return element_wise_shape(shape_a, self.arguments['shape'])[:2]


class Pow(Operator):
    def __init__(self):
        self.operator_sign = '**'
        self.inputs_count = 2

    def f(self, value_a, value_b):
        return numpy.power(value_a, value_b)

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),symbol_b)(Mul(),symbol_a(Pow(),symbol_b(Sub(),1))),
                lambda: forward(Mul(),symbol_a(Pow(),symbol_b))(Mul(),symbol_a(Log()))]

    def shape(self, shape_a, shape_b):
        return element_wise_shape(shape_a, shape_b)


class Log(Operator):
    def __init__(self):
        self.inputs_count = 1

    def f(self, value_a):
        return numpy.log(value_a)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),1)(Div(),symbol_a)]

    def shape(self, shape_a):
        return shape_a, ()


class Where(Operator):
    def __init__(self):
        self.inputs_count = 3

    def f(self, value_condition, value_a, value_b):
        return numpy.array(numpy.where(value_condition, value_a, value_b), dtype=float)

    def bprop(self, engine, symbol_forward, symbol_condition, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        
        return [lambda: constant(0),
                lambda: forward(Mul(),V(oprator=Where(),inputs=as_symbols([symbol_condition, forward, 0]))),
                lambda: forward(Mul(),V(oprator=Where(),inputs=as_symbols([symbol_condition, 0,forward])))]

    def shape(self, shape_condition, shape_a, shape_b):
        return element_wise_shape(shape_condition, shape_a, shape_b)


class Eq(Operator):
    def __init__(self):
        self.operator_sign = '=='
        self.inputs_count = 2

    def f(self, value_a, value_b):
        return numpy.equal(value_a, value_b)

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        return [lambda: V(operator=Where(),inputs=as_symbols([symbol_a(Eq(),symbol_b), forward, 0])),
                lambda: V(operator=Where(),inputs=as_symbols([symbol_a(Eq(),symbol_b), forward, 0]))]

    def shape(self, shape_a, shape_b):
        return element_wise_shape(shape_a, shape_b)


class NEq(Operator):
    def __init__(self):
        self.operator_sign = '!='
        self.inputs_count = 2

    def f(self, value_a, value_b):
        return numpy.not_equal(value_a, value_b)

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        return [lambda:V(operator=Where(),inputs=as_symbols([symbol_a(NEq(),symbol_b), forward,0])),
                lambda:V(operator=Where(),inputs=as_symbols([symbol_a(NEq(),symbol_b), forward,0])),]

    def shape(self, shape_a, shape_b):
        return element_wise_shape(shape_a, shape_b)


class Lt(Operator):
    def __init__(self):
        self.operator_sign = '<'
        self.inputs_count = 2

    def f(self, value_a, value_b):
        return numpy.less(value_a, value_b)

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        return [lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Lt(),symbol_b), forward, 0])),
                lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Lt(),symbol_b), forward, 0])),]

    def shape(self, shape_a, shape_b):
        return element_wise_shape(shape_a, shape_b)


class LEq(Operator):
    def __init__(self):
        self.operator_sign = '<='
        self.inputs_count = 2

    def f(self, value_a, value_b):
        return numpy.less_equal(value_a, value_b)

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        return [lambda:V(operator=Where(),inputs=as_symbols([symbol_a(LEq(),symbol_b), forward, 0])),
                lambda:V(operator=Where(),inputs=as_symbols([symbol_a(LEq(),symbol_b), forward, 0])),]

    def shape(self, shape_a, shape_b):
        return element_wise_shape(shape_a, shape_b)


class Gt(Operator):
    def __init__(self):
        self.operator_sign = '>'
        self.inputs_count = 2

    def f(self, value_a, value_b):
        return numpy.greater(value_a, value_b)

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        return [lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Gt(),symbol_b), forward, 0])),
                lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Gt(),symbol_b), forward, 0])),]

    def shape(self, shape_a, shape_b):
        return element_wise_shape(shape_a, shape_b)


class GEq(Operator):
    def __init__(self):
        self.operator_sign = '>='
        self.inputs_count = 2

    def f(self, value_a, value_b):
        return numpy.greater_equal(value_a, value_b)

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        return [lambda:V(operator=Where(),inputs=as_symbols([symbol_a(GEq(),symbol_b), forward, 0])),
                lambda:V(operator=Where(),inputs=as_symbols([symbol_a(GEq(),symbol_b), forward, 0])),]

    def shape(self, shape_a, shape_b):
        return element_wise_shape(shape_a, shape_b)


class Max(Operator):
    def __init__(self):
        self.inputs_count = 2

    def f(self, value_a, value_b):
        return numpy.maximum(value_a, value_b)

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        return [lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Gt(),symbol_b), forward, 0])),
                lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Lt(),symbol_b), forward, 0])),]

    def shape(self, shape_a, shape_b):
        return element_wise_shape(shape_a, shape_b)


class Min(Operator):
    def __init__(self):
        self.inputs_count = 2

    def f(self, value_a, value_b):
        return numpy.minimum(value_a, value_b)

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        return [lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Lt(),symbol_b), forward, 0])),
                lambda:V(operator=Where(),inputs=as_symbols([symbol_a(Gt(),symbol_b), forward, 0])),]

    def shape(self, shape_a, shape_b):
        return element_wise_shape(shape_a, shape_b)


class Sin(Operator):
    def __init__(self):
        self.inputs_count = 1

    def f(self, value_a):
        return numpy.sin(value_a)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),symbol_a(Cos()))]

    def shape(self, shape_a):
        return shape_a, ()


class Cos(Operator):
    def __init__(self):
        self.inputs_count = 1

    def f(self, value_a):
        return numpy.cos(value_a)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),symbol_a(Sin())(Neg()))]

    def shape(self, shape_a):
        return shape_a, ()


class Tan(Operator):
    def __init__(self):
        self.inputs_count = 1

    def f(self, value_a):
        return numpy.tan(value_a)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),1)(Div(),symbol_a(Cos())(Pow(),2))]

    def shape(self, shape_a):
        return shape_a, ()


class ArcSin(Operator):
    def __init__(self):
        self.inputs_count = 1

    def f(self, value_a):
        return numpy.arcsin(value_a)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),1)(Div(),constant(1)(Sub(),symbol_a(Pow(),2))(Pow(),0.5))]

    def shape(self, shape_a):
        return shape_a, ()


class ArcCos(Operator):
    def __init__(self):
        self.inputs_count = 1

    def f(self, value_a):
        return numpy.arccos(value_a)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),-1)(Div(),constant(1)(Sub(),symbol_a(Pow(),constant(2)))(Pow(),0.5))]

    def shape(self, shape_a):
        return shape_a, ()


class ArcTan(Operator):
    def __init__(self):
        self.inputs_count = 1

    def f(self, value_a):
        return numpy.arctan(value_a)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),1)(Div(),constant(1)(Plus(),symbol_a(Pow(),2)))]

    def shape(self, shape_a):
        return shape_a, ()


class Sinh(Operator):
    def __init__(self):
        self.inputs_count = 1

    def f(self, value_a):
        return numpy.sinh(value_a)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),symbol_a(Cosh()))]

    def shape(self, shape_a):
        return shape_a, ()


class Cosh(Operator):
    def __init__(self):
        self.inputs_count = 1

    def f(self, value_a):
        return numpy.cosh(value_a)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),symbol_a(Sinh()))]

    def shape(self, shape_a):
        return shape_a, ()


class Tanh(Operator):
    def __init__(self):
        self.inputs_count = 1

    def f(self, value_a):
        return numpy.tanh(value_a)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),constant(1)(Sub(),symbol_a(Tanh())(Pow(),2)))]

    def shape(self, shape_a):
        return shape_a, ()


class ArcSinh(Operator):
    def __init__(self):
        self.inputs_count = 1

    def f(self, value_a):
        return numpy.arcsinh(value_a)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),1)(Div(),symbol_a(Pow(),2)(Plus(),1)(Pow(),0.5))]

    def shape(self, shape_a):
        return shape_a, ()


class ArcCosh(Operator):
    def __init__(self):
        self.inputs_count = 1

    def f(self, value_a):
        return numpy.arccosh(value_a)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),1)(Div(),symbol_a(Pow(),2)(Sub(),1)(Pow(),0.5))]

    def shape(self, shape_a):
        return shape_a, ()


class ArcTanh(Operator):
    def __init__(self):
        self.inputs_count = 1

    def f(self, value_a):
        return numpy.arctanh(value_a)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),1)(Div(),constant(1)(Sub(),symbol_a(Pow(),2)))]

    def shape(self, shape_a):
        return shape_a, ()


class Exp(Operator):
    def __init__(self):
        self.inputs_count = 1

    def f(self, value_a):
        return numpy.exp(value_a)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Mul(),symbol_a(Exp()))]

    def shape(self, shape_a):
        return shape_a, ()


class SliceAssign(Operator):
    def __init__(self, slice_tuple):
        self.inputs_count = 2
        self.arguments = {'slice_tuple': slice_tuple}

    def f(self, value_a, value_b):
        value_a[self.arguments['slice_tuple']] = value_b
        return value_a

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        slice_tuple = self.arguments['slice_tuple']
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(SliceAssign(slice_tuple),numpy.zeros(engine.shape2(symbol_b))),
                lambda: forward[slice_tuple]]

    def shape(self, shape_a, shape_b):
        shape_select = slice_shape(shape_a, self.arguments['slice_tuple'])[0]
        shape_package = element_wise_shape(shape_select, shape_b)
        if shape_package[0] != shape_select:
            raise ValueError('Can not assign: {} to {} with {}'.format(shape_b, shape_a, self.arguments['slice_tuple']))
        return shape_a, (), shape_package[2]


class SliceSelect(Operator):
    def __init__(self, slice_tuple):
        self.inputs_count = 1
        self.arguments = {'slice_tuple': slice_tuple}

    def f(self, value_a):
        return value_a[self.arguments['slice_tuple']]

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        symbol_zero = constant(numpy.zeros(engine.shape2(symbol_a)))
        return [lambda: symbol_zero(SliceAssign(forward), self.arguments['slice_tuple'])]

    def shape(self, shape_a):
        return slice_shape(shape_a, self.arguments['slice_tuple'])


class Concatenate(Operator):
    def __init__(self, axis: int=0):
        self.inputs_count = 2
        self.arguments = {'axis': axis}

    def f(self, value_a, value_b):
        return numpy.concatenate((value_a, value_b), **self.arguments)

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        split_dimension = engine.shape2(symbol_a)[self.arguments['axis']]
        total_dimension = engine.shape2(symbol_forward)[self.arguments['axis']]
        return [lambda: forward[[slice(None)] * self.arguments['axis'] + [slice(0, split_dimension)]],
                lambda: forward[[slice(None)] * self.arguments['axis'] + [slice(split_dimension, total_dimension)]]]

    def shape(self, *shapes):
        return concatenate_shape(self.arguments['axis'], *shapes)


class Rotate90(Operator):
    def __init__(self, count: int=1, axes: tuple=None):
        self.inputs_count = 1
        self.arguments = {'count': count, 'axes': axes}

    def f(self, value_a):
        return numpy.rot90(value_a, k=self.arguments['count'], axes=self.arguments['axes'])

    def bprop(self, engine, symbol_forward, symbol_a, symbol_b):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Rotate90(-self.arguments['count'] & 3, self.arguments['axes']))]

    def shape(self, shape_a):
        return rotate90_shape(shape_a, **self.arguments)


class Flip(Operator):
    def __init__(self, axis: int):
        self.inputs_count = 1
        self.arguments = {'axis': axis}

    def f(self, value_a):
        return numpy.flip(value_a, self.arguments['axis'])

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Flip(self.arguments['axis']))]

    def shape(self, shape_a):
        return shape_a, ()


class Reshape(Operator):
    def __init__(self, shape):
        self.inputs_count = 1
        self.arguments = {'shape': shape}

    def f(self, value_a):
        return numpy.reshape(value_a, self.arguments['shape'])

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Reshape(engine.shape2(symbol_a)))]

    def shape(self, shape_a):
        return self.arguments['shape'], ()


class Spread(Operator):
    def __init__(self, position):
        self.inputs_count = 1
        self.arguments = {'position': position}

    def f(self, value_a):
        shape_a = value_a.shape
        spread_dimension = numpy.prod(shape_a[self.arguments['position']:])#reduce(lambda a, b: a * b, shape_a[self.arguments['position']:])
        new_shape = shape_a[:self.arguments['position']] + (spread_dimension,)
        return numpy.reshape(value_a, new_shape)

    def bprop(self, engine, symbol_forward, symbol_a):
        forward = engine.bprop(symbol_forward)
        return [lambda: forward(Reshape(engine.shape2(symbol_a)))]

    def shape(self, shape_a):
        spread_dimension = reduce(lambda a, b: a * b, shape_a[self.arguments['position']:])
        new_shape = shape_a[:self.arguments['position']] + (spread_dimension,)
        return new_shape, ()


# def negative(a):
#     return V(operator=Neg(), inputs=as_symbols([a]))
#  
#  
# def absolute(a):
#     return V(operator=Abs(), inputs=as_symbols([a]))
#  
#  
# def plus(a, b):
#     return V(operator=Plus(), inputs=as_symbols([a, b]))
#  
#  
# def subtract(a, b):
#     return V(operator=Sub(), inputs=as_symbols([a, b]))
#  
#  
# def multiply(a, b):
#     return V(operator=Mul(), inputs=as_symbols([a, b]))
#  
#  
# def divide(a, b):
#     return V(operator=Div(), inputs=as_symbols([a, b]))
#  
#  
# def matrix_multiply(a, b):
#     return V(operator=MM(), inputs=as_symbols([a, b]))
#  
#  
# def power(a, b):
#     return V(operator=Pow(), inputs=as_symbols([a, b]))
#  
#  
# def log(a):
#     return V(operator=Log(), inputs=as_symbols([a]))
#  
#  
# def transpose(a, axes: tuple=None):
#     return V(operator=T(axes), inputs=as_symbols([a]))
#  
#  
# def reduce_sum(a, axis: int=None, invariant=False):
#     return V(operator=ReduceSum(axis, invariant), inputs=as_symbols([a]))
#  
#  
# def reduce_mean(a, axis: int=None, invariant=False):
#     return V(operator=ReduceMean(axis, invariant), inputs=as_symbols([a]))
#  
#  
# def expand(a, axis: int):
#     return V(operator=Expand(axis), inputs=as_symbols([a]))
#  
#  
# def broadcast(a, shape):
#     return V(operator=Broadcast(shape), inputs=as_symbols([a]))
#  
#  
# def where(condition, a, b):
#     return V(operator=Where(), inputs=as_symbols([condition, a, b]))
#  
#  
# def equal(a, b):
#     return V(operator=Eq(), inputs=as_symbols([a, b]))
#  
#  
# def not_equal(a, b):
#     return V(operator=NEq(), inputs=as_symbols([a, b]))
#  
#  
# def less(a, b):
#     return V(operator=Lt(), inputs=as_symbols([a, b]))
#  
#  
# def less_equal(a, b):
#     return V(operator=LEq(), inputs=as_symbols([a, b]))
#  
#  
# def greater(a, b):
#     return V(operator=Gt(), inputs=as_symbols([a, b]))
#  
#  
# def greater_equal(a, b):
#     return V(operator=GEq(), inputs=as_symbols([a, b]))
#  
#  
# def maximum(a, b):
#     return V(operator=Max(), inputs=as_symbols([a, b]))
#  
#  
# def minimum(a, b):
#     return V(operator=Min(), inputs=as_symbols([a, b]))
#  
#  
# def sin(a):
#     return V(operator=Sin(), inputs=as_symbols([a]))
#  
#  
# def cos(a):
#     return V(operator=Cos(), inputs=as_symbols([a]))
#  
#  
# def tan(a):
#     return V(operator=Tan(), inputs=as_symbols([a]))
#  
#  
# def arcsin(a):
#     return V(operator=ArcSin(), inputs=as_symbols([a]))
#  
#  
# def arccos(a):
#     return V(operator=ArcCos(), inputs=as_symbols([a]))
#  
#  
# def arctan(a):
#     return V(operator=ArcTan(), inputs=as_symbols([a]))
#  
#  
# def sinh(a):
#     return V(operator=Sinh(), inputs=as_symbols([a]))
#  
#  
# def cosh(a):
#     return V(operator=Cosh(), inputs=as_symbols([a]))
#  
#  
# def tanh(a):
#     return V(operator=Tanh(), inputs=as_symbols([a]))
#  
#  
# def arcsinh(a):
#     return V(operator=ArcSinh(), inputs=as_symbols([a]))
#  
#  
# def arccosh(a):
#     return V(operator=ArcCosh(), inputs=as_symbols([a]))
#  
#  
# def arctanh(a):
#     return V(operator=ArcTanh(), inputs=as_symbols([a]))
#  
#  
# def exp(a):
#     return V(operator=Exp(), inputs=as_symbols([a]))
#  
#  
# def slice_assign(a, b, slice_tuple):
#     return V(operator=SliceAssign(slice_tuple), inputs=as_symbols([a, b]))
#  
#  
# def assign(a, b):
#     return slice_assign(a, b, slice(None))
#  
#  
# def slice_select(a, slice_tuple):
#     return V(operator=SliceSelect(slice_tuple), inputs=as_symbols([a]))
#  
#  
# def concatenate(a, b):
#     return V(operator=Concatenate(), inputs=as_symbols([a, b]))
#  
#  
# def rotate90(a, count, axes):
#     return V(operator=Rotate90(count, axes), inputs=as_symbols([a]))
#  
#  
# def flip(a, axis):
#     return V(operator=Flip(axis), inputs=as_symbols([a]))
#  
# def reshape(a, shape):
#     return V(operator=Reshape(shape), inputs=as_symbols([a]))
#  
#  
# def spread(a, position):
#     return V(operator=Spread(position), inputs=as_symbols([a]))
def where(a,b,c):
    return V(operator=Where(),inputs=as_symbols([a,b,c]))


from .tensor import V,constant
from .interface import Optimizer

class GradientDescentOptimizer(Optimizer):
    def __init__(self, rate: float, consistent: bool=True):
        self.__rate = rate
        self.__consistent = consistent
        self.__gradient_engine=constant(1)
    def __repr__(self):
        return '{}(rate={}, consistent={})'.format(self.__class__.__name__, self.__rate, self.__consistent)

    def optimize(self, engine , calculate_function):
        variables = engine.variables
        for variable in variables:
            value_cache = self.__gradient_engine.value_cache
            self.__gradient_engine = engine.bprop(variable)
            if self.__gradient_engine:self.__gradient_engine.bind=engine.bind
            else :self.__gradient_engine=constant(1)
            if self.__consistent:
                self.__gradient_engine.value_cache = value_cache
            variable.value = calculate_function(variable.value, self.__rate * self.__gradient_engine.val())
        engine.modified()
        self.__gradient_engine.modified()


class MomentumOptimizer(Optimizer):
    def __init__(self, rate: float, factor: float=0, consistent: bool=True):
        self.__rate = rate
        self.__factor = factor
        self.__consistent = consistent
        self.__old_gradient_map = {}
        self.__gradient_engine = V()

    def __repr__(self):
        return '{}(rate={}, consistent={})'.format(self.__class__.__name__, self.__rate, self.__consistent)

    def optimize(self, engine, calculate_function):
        variables = engine.variables
        for variable in variables:
            value_cache = self.__gradient_engine.value_cache
            self.__gradient_engine = engine.bprop(variable)
            self.__gradient_engine.set_bind(engine.get_bind())
            if self.__consistent:
                self.__gradient_engine.value_cache = value_cache
            momentum = self.__gradient_engine.val() + self.__factor * self.__old_gradient_map.get(variable, 0)
            self.__old_gradient_map[variable] = momentum
            variable.value = calculate_function(variable.value, self.__rate * momentum)
        engine.modified()
        self.__gradient_engine.modified()


class AdaptiveGradientOptimizer(Optimizer):
    def __init__(self, rate: float, consistent: bool=True):
        self.__rate = rate
        self.__consistent = consistent
        self.__gradient_engine = V()
        self.__accumulate_gradient_map = {}

    def __repr__(self):
        return '{}(rate={}, consistent={})'.format(self.__class__.__name__, self.__rate, self.__consistent)

    def optimize(self, engine, calculate_function):
        variables = engine.variables
        for variable in variables:
            value_cache = self.__gradient_engine.value_cache
            self.__gradient_engine = engine.bprop(variable)
            self.__gradient_engine.set_bind(engine.get_bind())
            if self.__consistent:
                self.__gradient_engine.value_cache = value_cache
            current_gradient = self.__gradient_engine.val()
            self.__accumulate_gradient_map.setdefault(variable, 0)
            self.__accumulate_gradient_map[variable] += current_gradient ** 2
            regularization_value = current_gradient / (self.__accumulate_gradient_map[variable] + 1e-8) ** 0.5
            variable.value = calculate_function(variable.value, self.__rate * regularization_value)
        engine.modified()
        self.__gradient_engine.modified()


class AdaptiveDeltaOptimizer(Optimizer):
    def __init__(self, decay: float, consistent: bool=True):
        self.__decay = decay
        self.__consistent = consistent
        self.__gradient_engine = V()
        self.__accumulate_gradient_map = {}
        self.__expectation_map = {}

    def __repr__(self):
        return '{}(decay={}, consistent={})'.format(self.__class__.__name__, self.__decay, self.__consistent)

    def optimize(self, engine, calculate_function):
        variables = engine.variables
        for variable in variables:
            value_cache = self.__gradient_engine.value_cache
            self.__gradient_engine = engine.bprop(variable)
            self.__gradient_engine.set_bind(engine.get_bind())
            if self.__consistent:
                self.__gradient_engine.value_cache = value_cache
            current_gradient = self.__gradient_engine.val()
            self.__accumulate_gradient_map.setdefault(variable, 0)
            self.__expectation_map.setdefault(variable, 0)
            self.__accumulate_gradient_map[variable] = self.__decay * self.__accumulate_gradient_map[variable] + (1 - self.__decay) * current_gradient ** 2
            delta = (self.__expectation_map[variable] + 1e-8) ** 0.5 / (self.__accumulate_gradient_map[variable] + 1e-8) ** 0.5 * current_gradient
            self.__expectation_map[variable] = self.__decay * self.__expectation_map[variable] + (1 - self.__decay) * delta ** 2
            variable.value = calculate_function(variable.value, delta)
        engine.modified()
        self.__gradient_engine.modified()


class RootMeanSquarePropOptimizer(Optimizer):
    def __init__(self, rate: float, consistent: bool=True):
        self.__rate = rate
        self.__consistent = consistent
        self.__gradient_engine = V()
        self.__mean_map = {}
        self.__step = 1

    def __repr__(self):
        return '{}(rate={}, consistent={})'.format(self.__class__.__name__, self.__rate, self.__consistent)

    def optimize(self, engine, calculate_function):
        variables = engine.variables
        for variable in variables:
            value_cache = self.__gradient_engine.value_cache
            self.__gradient_engine = engine.bprop(variable)
            self.__gradient_engine.set_bind(engine.get_bind())
            if self.__consistent:
                self.__gradient_engine.value_cache = value_cache
            current_gradient = self.__gradient_engine.val()
            self.__mean_map.setdefault(variable, 0)
            self.__mean_map[variable] *= (self.__step - 1) / self.__step
            self.__mean_map[variable] += current_gradient ** 2 / self.__step
            self.__step += 1
            regularization_value = current_gradient / (self.__mean_map[variable] + 1e-8) ** 0.5
            variable.value = calculate_function(variable.value, self.__rate * regularization_value)
        engine.modified()
        self.__gradient_engine.modified()


class AdaptiveMomentEstimationOptimizer(Optimizer):
    def __init__(self, rate: float, decay: float=0.9, square_decay: float=0.999, consistent: bool=True):
        self.__rate = rate
        self.__decay = decay
        self.__square_decay = square_decay
        self.__consistent = consistent
        self.__gradient_engine = V()
        self.__estimation_map = {}
        self.__square_estimation_map = {}
        self.__step = 1

    def __repr__(self):
        return '{}(rate={}, decay={}, square_decay={}, consistent={})'.format(self.__class__.__name__, self.__rate, self.__decay, self.__square_decay, self.__consistent)

    def optimize(self, engine, calculate_function):
        
        variables = engine.variables
        for variable in variables:
            
            value_cache = self.__gradient_engine.value_cache
            self.__gradient_engine = engine.bprop(variable)
            self.__gradient_engine.bind=engine.bind
            if self.__consistent:
                self.__gradient_engine.value_cache = value_cache
            current_gradient = self.__gradient_engine.val()
            self.__estimation_map.setdefault(variable, 0)
            self.__square_estimation_map.setdefault(variable, 0)
            self.__estimation_map[variable] = self.__decay * self.__estimation_map[variable] + (1 - self.__decay) * current_gradient
            self.__square_estimation_map[variable] = self.__square_decay * self.__square_estimation_map[variable] + (1 - self.__square_decay) * current_gradient ** 2
            estimation = self.__estimation_map[variable] / (1 - self.__decay ** self.__step)
            square_estimation = self.__square_estimation_map[variable] / (1 - self.__square_decay ** self.__step)
            self.__step += 1
            regularization_value = estimation / (square_estimation + 1e-8) ** 0.5
            variable.value = calculate_function(variable.value, self.__rate * regularization_value)
        engine.modified()
        self.__gradient_engine.modified()
        



from ..sheng import V, ReduceSum,Plus,Div,Max,Exp,Tanh,Neg,constant,Log,Abs
from ..yu import xavier_initialization, he_initialization, bias_initialization


class ActivationLayer:
    def activation_layer(self):
        return self

    def activation_function(self, *args, **kwargs):
        raise ValueError("this is abstract")

    @staticmethod
    def weight_initialization(shape):
        weight = xavier_initialization(shape)
        return weight

    @staticmethod
    def bias_initialization(shape):
        bias = bias_initialization(shape)
        return bias


class RectifiedLinearUnits(ActivationLayer):
    @staticmethod
    def activation_function(input_symbol: V):
        output_symbol = input_symbol(Max(), 0)
        return output_symbol

    @staticmethod
    def weight_initialization(shape):
        weight = he_initialization(shape)
        return weight


class SoftMax(ActivationLayer):
    @staticmethod
    def activation_function(input_symbol: V):
        exp_symbol = input_symbol(Exp())
        output_symbol = exp_symbol(Div(),exp_symbol(ReduceSum()))
        return output_symbol


class HyperbolicTangent(ActivationLayer):
    @staticmethod
    def activation_function(input_symbol: V):
        output_symbol = input_symbol(Tanh())
        return output_symbol


class Sigmoid(ActivationLayer):
    @staticmethod
    def activation_function(input_symbol: V):
        output_symbol = constant(1)(Div(),input_symbol(Neg())(Exp())(Plus(),1))
        return output_symbol


class SoftPlus(ActivationLayer):
    @staticmethod
    def activation_function(input_symbol: V):
        output_symbol = input_symbol(Exp())(Plus(),1)(Log())
        return output_symbol


class SoftSign(ActivationLayer):
    @staticmethod
    def activation_function(input_symbol: V):
        output_symbol = input_symbol(Div(),input_symbol(Abs())(Plus(),1))
        return output_symbol


relu = RectifiedLinearUnits.activation_function
softmax = SoftMax.activation_function
sigmoid = Sigmoid.activation_function
softplus = SoftPlus.activation_function
softsign = SoftSign.activation_function


activation_map = {
    'relu': RectifiedLinearUnits,
    'softmax': SoftMax,
    'tanh': HyperbolicTangent,
    'sigmoid': Sigmoid,
    'softplus': SoftPlus,
    'softsign': SoftSign,
}


def register_activation(name: str, activation: ActivationLayer):
    activation_map[name.lower()] = activation


class Activation:
    def __init__(self, name: str, *args, **kwargs):
        self.__name = name.lower()
        self.__activation = None
        if self.__name in activation_map:
            self.__activation = activation_map[self.__name](*args, **kwargs)
        else:
            raise ValueError('No such activation: {}'.format(name))

    def activation_layer(self):
        return self.__activation



from ..sheng import V, variable,Spread, ReduceMean,numpy,MM,Plus,Sub,Pow,Div,Mul
from ..yu.initialization import xavier_initialization, bias_initialization


class ConnectionLayer:
    def __init__(self, output_shape=None, input_shape=None):
        self._output_shape = output_shape
        self._input_shape = input_shape

    def get_output_shape(self):
        return self._output_shape

    def set_output_shape(self, output_shape):
        self._output_shape = output_shape

    output_shape = property(get_output_shape, set_output_shape)

    def get_input_shape(self):
        return self._input_shape

    def set_input_shape(self, input_shape):
        self._input_shape = input_shape

    input_shape = property(get_input_shape, set_input_shape)

    def connection_layer(self):
        return self

    def connection(self, input_symbol: V):
        raise ValueError("this is abstract")

    def variables(self):
        raise ValueError("this is abstract")

    def weights(self):
        return []

    def biases(self):
        return []


class Dense(ConnectionLayer):
    def __init__(self, output_dimension: int, input_dimension: int=None):
        ConnectionLayer.__init__(self, output_dimension, input_dimension)
        self.__weight = None
        self.__bias = None
        self.__output_symbol = None

    def connection(self, input_symbol: V):
        if self.__output_symbol is None:
            if not isinstance(self._input_shape, int):
                input_symbol = input_symbol(Spread(1))
                self._input_shape = input_symbol.operator.shape((None,) + self._input_shape)[0][1]
            weight, bias = self.variables()
            self.__output_symbol = input_symbol(MM(),weight)(Plus(),bias)
        return self.__output_symbol

    def variables(self):
        if self.__weight is None:
            self.__weight = variable(xavier_initialization((self._input_shape, self._output_shape)))
        if self.__bias is None:
            self.__bias = variable(bias_initialization((1, self._output_shape)))
        return [self.__weight, self.__bias]

    def weights(self):
        return [self.__weight]

    def biases(self):
        return [self.__bias]


class BatchNormalization(ConnectionLayer):
    def __init__(self, input_shape=None):
        ConnectionLayer.__init__(self, input_shape, input_shape)
        self.__scale = None
        self.__shift = None
        self.__output_symbol = None
        self.__input_mean = None
        self.__input_variance = None

    def connection(self, input_symbol: V):
        if self.__output_symbol is None:
            self.__input_mean = input_symbol(ReduceMean(axis=0))
            self.__input_variance = input_symbol(Sub(),self.__input_mean)(Pow(),2)(ReduceMean(axis=0))
            input_normalize = input_symbol (Sub(),self.__input_mean)(Div(),self.__input_variance(Plus(),1e-8)(Pow(),0.5))
            scale, shift = self.variables()
            self.__output_symbol = scale(Mul(),input_normalize)(Plus(),shift)
            self._output_shape = self._input_shape
        return self.__output_symbol

    def variables(self):
        if self.__scale is None:
            self.__scale = variable(numpy.ones(self._input_shape))
        if self.__shift is None:
            self.__shift = variable(numpy.zeros(self._input_shape))
        return [self.__scale, self.__shift]

    def normalization_symbol(self):
        return [self.__input_mean, self.__input_variance]


connection_map = {
    'dense': Dense,
    'batch normalization': BatchNormalization,
}


def register_connection(name: str, connection: ConnectionLayer):
    connection_map[name.lower()] = connection


class Connection:
    def __init__(self, name: str, *args, **kwargs):
        self.__name = name.lower()
        self.__connection = None
        if self.__name in connection_map:
            self.__connection = connection_map[self.__name](*args, **kwargs)
        else:
            raise ValueError('No such connection: {}'.format(name))

    def connection_layer(self):
        return self.__connection



import collections
from ..sheng import Operator,numpy,Optimizer,VCategory,V, constant, placeholder,Plus
from ..sheng.optim import GradientDescentOptimizer,MomentumOptimizer, \
    AdaptiveDeltaOptimizer,AdaptiveGradientOptimizer, \
    AdaptiveMomentEstimationOptimizer,RootMeanSquarePropOptimizer
from .loss import LossLayer, Loss
from .regularization import RegularizationLayer, Regularization
from .connection import ConnectionLayer, Connection
from .activation import ActivationLayer, Activation
from .cnnlayer import Convolution, Pooling, Unpooling
from .plugin import Plugin, default_plugin

optimizer_map = {
    'sgd': GradientDescentOptimizer,
    'mmt': MomentumOptimizer,
    'adag': AdaptiveGradientOptimizer,
    'adad': AdaptiveDeltaOptimizer,
    'rmsprop': RootMeanSquarePropOptimizer,
    'adam': AdaptiveMomentEstimationOptimizer,
}


def register_optimizer(name: str, optimizer: Optimizer):
    optimizer_map[name.lower()] = optimizer


class Model:
    def __init__(self):
        self.epoch = None
        self.epochs = None
        self.iteration = None
        self.iterations = None
        self.batch_size = None
        self.engine = V()
        self.predict_engine = V()
        self.__layer = []
        self.__layer_name_map = collections.OrderedDict()
        self.__layer_stack = []
        self.__input_symbol = placeholder(name='InputSymbol')
        self.__current_symbol = self.__input_symbol
        self.__current_output = None
        self.__variables = []
        self.__layer_weights = {}
        self.__data = None
        self.__optimizer = None
        self.__loss = None
        self.__regularization_term = None
        self.__plugin = collections.OrderedDict()
        self.load_default_plugin()

    def __valid_current_output(self):
        if self.__current_output is None:
            raise ValueError('Current output is None.')
        else:
            return self.__current_output

    def get_layer(self, name: str):
        if name in self.__layer_name_map:
            return self.__layer_name_map[name]
        else:
            raise ValueError('No such layer in Model named: {}'.format(name))

    def layer_name_map(self):
        return self.__layer_name_map

    def layer_stack(self):
        return self.__layer_stack

    def add(self, layer, name=None):
        if isinstance(layer, collections.Iterable):
            for i, l in enumerate(layer):
                if name is not None and i < len(name):
                    self.__add(l, name[i])
                else:
                    self.__add(l)
        else:
            self.__add(layer, name)

    def __add(self, layer, name: str=None):
        self.__layer.append(layer)
        if name is None:
            name = 'layer_{}'.format(len(self.__layer_stack))
        if name in self.__layer_name_map:
            raise ValueError('Layer name has contained in Model: {}'.format(name))
        else:
            self.__layer_name_map[name] = layer
        self.__layer_stack.append(layer)
        if isinstance(layer, Operator):
            self.__add_operator(layer, name)
        elif isinstance(layer, ConnectionLayer):
            self.__add_connection(layer, name)
        elif isinstance(layer, Connection):
            self.__add_connection(layer.connection_layer(), name)
        elif isinstance(layer, Convolution):
            self.__add_connection(layer.convolution_layer(), name)
        elif isinstance(layer, Pooling):
            self.__add_connection(layer.pooling_layer(), name)
        elif isinstance(layer, Unpooling):
            self.__add_connection(layer.unpooling_layer(), name)
        elif isinstance(layer, ActivationLayer):
            self.__add_activation(layer, name)
        elif isinstance(layer, Activation):
            self.__add_activation(layer.activation_layer(), name)
        else:
            raise ValueError('Invalid get_layer type: {}'. format(type(layer)))

    def __add_operator(self, layer: Operator, name: str=None):
        self.__current_symbol = V(operator=layer, inputs=[self.__current_symbol], category=VCategory.operator)
        self.__current_output = layer.shape(self.__current_output)
        self.__layer_weights[name] = []

    def __add_connection(self, layer: ConnectionLayer, name: str=None):
        if layer.input_shape is None:
            layer.input_shape = self.__valid_current_output()
        self.__current_symbol = layer.connection(self.__current_symbol)
        self.__current_output = layer.output_shape
        for v in layer.variables():
            self.__variables.append(v)
        self.__layer_weights[name] = layer.weights()

    def __add_activation(self, layer: ActivationLayer, name: str=None):
        self.__current_symbol = layer.activation_function(self.__current_symbol)
        previous_layer = self.__layer_stack[-2]
        if isinstance(previous_layer, Connection) or isinstance(previous_layer, ConnectionLayer):
            previous_layer = previous_layer.connection_layer()
            for weight in previous_layer.weights():
                weight.value = layer.weight_initialization(weight.value.shape)
            for bias in previous_layer.biases():
                bias.value = layer.bias_initialization(bias.value.shape)
        self.__layer_weights[name] = []

    def get_symbol(self):
        return self.__current_symbol

    def optimizer(self, optimizer_object, *args, **kwargs):
        if isinstance(optimizer_object, str):
            name = optimizer_object.lower()
            if name in optimizer_map:
                self.__optimizer = optimizer_map[name](*args, **kwargs)
            else:
                raise ValueError('No such optimizer: {}'.format(name))
        elif isinstance(optimizer_object, Optimizer):
            self.__optimizer = optimizer_object
        else:
            raise ValueError('Invalid optimizer type: {}'.format(type(optimizer_object)))

    def loss(self, loss_object, *args, **kwargs):
        if isinstance(loss_object, str):
            self.__loss = Loss(loss_object, *args, **kwargs).loss_layer()
        elif isinstance(loss_object, LossLayer):
            self.__loss = loss_object
        elif isinstance(loss_object, Loss):
            self.__loss = loss_object.loss_layer()
        else:
            raise ValueError('Invalid loss type: {}'.format(type(loss_object)))

    def regularization(self, regularization_object, decay: float, name=None, *args, **kwargs):
        regularization_weights = set()
        if name is None:
            for _, weights in self.__layer_weights.items():
                regularization_weights |= set(weights)
        else:
            if isinstance(name, str):
                name = [name]
            if isinstance(name, collections.Iterable):
                for each_name in name:
                    if each_name in self.__layer_weights:
                        regularization_weights |= set(self.__layer_weights[each_name])
                    else:
                        raise ValueError('No such layer in Model named: {}'.format(each_name))
            else:
                ValueError('Invalid name type: {}'.format(type(name)))
        if isinstance(regularization_object, str):
            regularization_function = Regularization(regularization_object, *args, **kwargs).regularization_layer().regularization_term
        elif isinstance(regularization_object, RegularizationLayer):
            regularization_function = regularization_object.regularization_term
        elif isinstance(regularization_object, Regularization):
            regularization_function = regularization_object.regularization_layer().regularization_term
        else:
            raise ValueError('Invalid regularization type: {}'.format(type(regularization_object)))
        for weight in regularization_weights:
            self.__add_regularization_term(regularization_function(weight, decay))

    def __add_regularization_term(self, regularization_term):
        if self.__regularization_term is None:
            self.__regularization_term = regularization_term
        else:
            self.__regularization_term = self.__regularization_term(Plus(),regularization_term)

    def train(self, data, target, epochs: int=10000, batch_size: int=0):
        data = numpy.array(data)
        target = numpy.array(target)
        self.epochs = epochs
        if data.shape[0] != target.shape[0]:
            raise ValueError('Data dimension not match target dimension: {} {}'.format(data.shape[0], target.shape[0]))
        data_scale = data.shape[0]
        target_symbol = None
        if batch_size != 0:
            target_symbol = placeholder()
            self.engine = self.__loss.loss_function(self.__current_symbol, target_symbol)
        else:
            self.engine = self.__loss.loss_function(self.__current_symbol, constant(target))
            self.engine.set_bind({self.__input_symbol: data})
        if self.__regularization_term is not None:
            self.engine =self.engine(Plus(),self.__regularization_term)
        self.engine.variables = self.__variables
        try:
            self.iteration = 0
            iterations = [0] if batch_size == 0 else range(0, data_scale, batch_size)
            self.iterations = self.epochs * len(iterations)
            self.run_plugin('begin_training')
            for epoch in range(self.epochs):
                self.epoch = epoch + 1
                self.run_plugin('begin_epoch')
                for i in iterations:
                    if batch_size != 0:
                        self.engine.set_bind({
                            self.__input_symbol: data[i: min([i + batch_size, data_scale])],
                            target_symbol: target[i: min([i + batch_size, data_scale])]
                        })
                        
                    self.run_plugin('begin_iteration')
                    self.__optimizer.minimize(self.engine)
                    self.iteration += 1
                    self.run_plugin('end_iteration')
                self.run_plugin('end_epoch')
        except KeyboardInterrupt:
            print('<Keyboard Interrupt>')
        self.run_plugin('end_training')

    def predict(self, data):
        self.predict_engine = self.__current_symbol
        self.predict_engine.set_bind({self.__input_symbol: data})
        self.run_plugin('begin_predict')
        predict_data = self.predict_engine.val()
        self.run_plugin('end_predict')
        return predict_data

    def load_default_plugin(self):
        for name, plugin, enable in default_plugin:
            plugin.enable = enable
            self.add_plugin(name, plugin)

    def add_plugin(self, name: str, plugin: Plugin):
        self.__plugin[name.lower()] = plugin
        plugin.bind_network(self)

    def run_plugin(self, stage: str):
        for _, plugin in self.__plugin.items():
            if plugin.enable:
                getattr(plugin, stage)()

    def plugin(self, name: str):
        if name.lower() in self.__plugin:
            return self.__plugin[name.lower()]
        else:
            raise ValueError('No such plugin: {}'.format(name))



from functools import reduce
from ..interface import element_wise_shape,numpy
from ...yu import array_index_traversal, multi_range
from enum import Enum
class ConvolutionMode(Enum):
    valid = 0
    full = 1


convolution_map = {
    'valid': ConvolutionMode.valid,
    'full': ConvolutionMode.full,
}


def __get_convolution_mode_string(mode):
    if isinstance(mode, str):
        if mode in convolution_map:
            return mode
        else:
            raise ValueError('No such convolution mode: {}'.format(mode))
    elif isinstance(mode, ConvolutionMode):
        return mode.name
    else:
        raise ValueError('Invalid mode type: {}'.format(type(mode)))


def basic_convolution_shape(shape_data, shape_kernel, dimension: int, mode: str):
    if mode == 'valid':
        return tuple(shape_data[i] - shape_kernel[i] + 1 for i in range(-dimension, 0))
    elif mode == 'full':
        return tuple(shape_data[i] + shape_kernel[i] - 1 for i in range(-dimension, 0))
    else:
        raise ValueError('Invalid convolution mode: {}'.format(mode))

def __compute_valid_convolution_nd(data, kernel, dimension: int):
    convolution_shape = tuple(data.shape[i] - kernel.shape[i] + 1 for i in range(-1, -dimension - 1, -1))
    list_dimension = reduce(lambda a, b: a * b, convolution_shape)
    data_prefix = data.shape[:-dimension]
    kernel_flat = kernel.ravel()
    data_flat = numpy.zeros(data_prefix + (list_dimension, len(kernel_flat)))
    for i in range(list_dimension):
        tensor_slice_start = [0] * len(kernel.shape)
        tensor_slice = [slice(None)] * len(data.shape)
        tensor_slice_start[-1] = i
        for r in range(-1, -len(kernel.shape) - 1, -1):
            dimension_scale = data.shape[r] - kernel.shape[r] + 1
            if tensor_slice_start[r] >= dimension_scale:
                tensor_slice_start[r + 1] = tensor_slice_start[r] // dimension_scale
                tensor_slice_start[r] %= dimension_scale
            tensor_slice[r] = slice(tensor_slice_start[r], tensor_slice_start[r] + kernel.shape[r])
        sub_convolution_index = (slice(None),) * (len(data.shape) - dimension) + tuple([i, slice(None)])
        data_flat[tuple(sub_convolution_index)] = data[tuple(tensor_slice)].reshape(data_prefix + (reduce(lambda a, b: a * b, kernel.shape),))
    convolution_flat = numpy.matmul(data_flat, numpy.flip(kernel_flat, axis=0))
    convolution_nd = convolution_flat.reshape(data_prefix + convolution_shape)
    return convolution_nd

def __compute_convolution_nd(data, kernel, dimension: int, mode: str):
    mode_string = __get_convolution_mode_string(mode)
    for i in range(dimension):
        if data.shape[i] < kernel.shape[i]:
            raise ValueError('Data shape smaller than kernel shape: {} {}'.format(data.shape, kernel.shape))
    if mode_string == 'valid':
        return __compute_valid_convolution_nd(data, kernel, dimension)
    elif mode_string == 'full':
        data_prefix = data.shape[:-dimension]
        expand_data = numpy.zeros(data_prefix + tuple(data.shape[i] + (kernel.shape[i] - 1) * 2 for i in range(dimension)))
        padding = tuple(kernel.shape[i] - 1 for i in range(dimension))
        expand_data[tuple(slice(None) for _ in data_prefix) + tuple(slice(padding[i], padding[i] + data.shape[i]) for i in range(dimension))] = data
        return __compute_valid_convolution_nd(expand_data, kernel, dimension)
    else:
        raise ValueError('Never reached.')


def compute_convolution_nd(data, kernel, dimension: int, mode=ConvolutionMode.valid, element_wise: bool=False):
    mode_string = __get_convolution_mode_string(mode)
    result = []
    data_prefix_shape = data.shape[:-dimension]
    kernel_prefix_shape = kernel.shape[:-dimension]
    if element_wise:
        final_shape = element_wise_shape(data_prefix_shape, kernel_prefix_shape)[0]
        data = numpy.broadcast_to(data, final_shape + data.shape[-2:])
        kernel = numpy.broadcast_to(kernel, final_shape + kernel.shape[-2:])
        if final_shape:
            for index in array_index_traversal(final_shape):
                result.append(__compute_convolution_nd(data[index], kernel[index], dimension, mode_string))
            return numpy.array(result).reshape(final_shape + result[0].shape)
        else:
            return __compute_convolution_nd(data, kernel, dimension, mode_string)
    else:
        if kernel_prefix_shape:
            final_shape = data_prefix_shape + kernel_prefix_shape + basic_convolution_shape(data.shape[-dimension:], kernel.shape[-dimension:], dimension, mode_string)
            result = numpy.zeros(final_shape)
            for kernel_index in array_index_traversal(kernel_prefix_shape):
                sub_result_index = tuple(slice(None) for _ in data_prefix_shape) + kernel_index + tuple(slice(None) for _ in range(dimension))
                result[sub_result_index] = __compute_convolution_nd(data, kernel[kernel_index], dimension, mode_string)
            return result
        else:
            return __compute_convolution_nd(data, kernel, dimension, mode_string)

def __compute_max_pooling_nd(data, size, step, dimension: int, reference=None):
    for i in range(dimension):
        if data.shape[i] < size[i]:
            raise ValueError('Data shape smaller than size: {} {}'.format(data.shape, size))
    pooling_array = []
    pooling_grid = [range(0, data.shape[i] - size[i] + 1, step[i]) for i in range(dimension)]
    for index in multi_range(pooling_grid):
        sub_slice = [slice(index[i], index[i] + size[i]) for i in range(dimension)]
        if reference is None:
            pooling_array.append(numpy.max(data[tuple(sub_slice)]))
        else:
            max_index = numpy.argmax(reference[sub_slice])
            sub_data = data[sub_slice]
            pooling_array.append(sub_data[numpy.unravel_index(max_index, sub_data.shape)])
    
    return numpy.array(pooling_array).reshape([len(g) for g in pooling_grid])

def compute_max_pooling_nd(data, size, step, dimension: int, reference=None):
    result = []
    data_prefix_shape = data.shape[:-dimension]
    if data_prefix_shape:
        for key in array_index_traversal(data_prefix_shape):
            if reference is None:
                result.append(__compute_max_pooling_nd(data[key], size, step, dimension))
            else:
                result.append(__compute_max_pooling_nd(data[key], size, step, dimension, reference[key]))
        return numpy.array(result).reshape(data_prefix_shape + result[0].shape)
    else:
        return __compute_max_pooling_nd(data, size, step, dimension)


def __compute_max_unpooling_nd(data, pooling, size, step, dimension: int):
    for i in range(dimension):
        if data.shape[i] < size[i]:
            raise ValueError('Data shape smaller than size: {} {}'.format(data.shape, size))
    unpooling_array = numpy.zeros(data.shape)
    unpooling_grid = [range(0, data.shape[i] - size[i] + 1, step[i]) for i in range(dimension)]
    for n, index in enumerate(multi_range(unpooling_grid)):
        sub_slice = tuple([slice(index[i], index[i] + size[i]) for i in range(dimension)])
        max_index = numpy.argmax(data[sub_slice])
        sub_unpooling_array = unpooling_array[sub_slice]
        sub_unpooling_array[numpy.unravel_index(max_index, sub_unpooling_array.shape)] = pooling[numpy.unravel_index(n, pooling.shape)]
    return unpooling_array


def compute_max_unpooling_nd(data, pooling, size, step, dimension: int):
    result = []
    data_prefix_shape = data.shape[:-dimension]
    kernel_prefix_shape = pooling.shape[:-dimension]
    final_shape = element_wise_shape(data_prefix_shape, kernel_prefix_shape)[0]
    data = numpy.broadcast_to(data, final_shape + data.shape[-dimension:])
    pooling = numpy.broadcast_to(pooling, final_shape + pooling.shape[-dimension:])
    if final_shape:
        for key in array_index_traversal(final_shape):
            result.append(__compute_max_unpooling_nd(data[key], pooling[key], size, step, dimension))
        return numpy.array(result).reshape(final_shape + result[0].shape)
    else:
        return __compute_max_unpooling_nd(data, pooling, size, step, dimension)


def __compute_average_pooling_nd(data, size, step, dimension: int):
    for i in range(dimension):
        if data.shape[i] < size[i]:
            raise ValueError('Data shape smaller than size: {} {}'.format(data.shape, size))
    pooling_array = []
    pooling_grid = [range(0, data.shape[i] - size[i] + 1, step[i]) for i in range(dimension)]
    for index in multi_range(pooling_grid):
        pooling_array.append(numpy.mean(data[tuple([slice(index[i], index[i] + size[i]) for i in range(dimension)])]))
    return numpy.array(pooling_array).reshape([len(g) for g in pooling_grid])


def compute_average_pooling_nd(data, size, step, dimension: int):
    result = []
    data_prefix_shape = data.shape[:-dimension]
    if data_prefix_shape:
        for key in array_index_traversal(data_prefix_shape):
            result.append(__compute_average_pooling_nd(data[key], size, step, dimension))
        return numpy.array(result).reshape(data_prefix_shape + result[0].shape)
    else:
        return __compute_average_pooling_nd(data, size, step, dimension)


def __compute_average_unpooling_nd(pooling, size, step, dimension: int, unpooling_size=None):
    if unpooling_size is None:
        unpooling_array = numpy.zeros([size[i] + (pooling.shape[i] - 1) * step[i] for i in range(dimension)])
    else:
        unpooling_array = numpy.zeros(unpooling_size)
    unpooling_grid = [range(0, unpooling_array.shape[i] - size[i] + 1, step[i]) for i in range(dimension)]
    for n, index in enumerate(multi_range(unpooling_grid)):
        sub_slice = tuple(slice(index[i], index[i] + size[i]) for i in range(dimension))
        unpooling_array[sub_slice] += pooling[numpy.unravel_index(n, pooling.shape)]
    return unpooling_array


def compute_average_unpooling_nd(pooling, size, step, dimension: int, unpooling_size=None):
    result = []
    data_prefix_shape = pooling.shape[:-dimension]
    if data_prefix_shape:
        for key in array_index_traversal(data_prefix_shape):
            result.append(__compute_average_unpooling_nd(pooling[key], size, step, dimension, unpooling_size))
        return numpy.array(result).reshape(data_prefix_shape + result[0].shape)
    else:
        return __compute_average_unpooling_nd(pooling, size, step, dimension, unpooling_size)



from ..sheng import V,Exp,Mul,Neg,Plus,Log,ReduceSum,ReduceMean,Div,Eq,where,Max,Pow,Sub,constant
from ..yu import generate_label_matrix


class LossCategory:
    classification = 0
    regression = 1


class LossLayer:
    loss_type = None

    def loss_function(self, *args, **kwargs):
        raise ValueError("this is abstract")

    
class SoftMaxLoss(LossLayer):
    loss_type = LossCategory.classification

    @staticmethod
    def loss_function(input_symbol: V, label_symbol: V):
        exp_symbol = input_symbol(Exp())
        softmax_value = label_symbol(Mul(),exp_symbol)(ReduceSum(axis=1))(Div(),exp_symbol(ReduceSum(axis=1)))
        loss = softmax_value(Log())(Neg())(ReduceMean())
        return loss


class SVMLoss(LossLayer):
    loss_type = LossCategory.classification

    @staticmethod
    def loss_function(input_symbol: V, label_symbol: V):
        dimension = label_symbol.shape[1]
        label_symbol = label_symbol(Mul(),-(dimension - 1))
        label_symbol = where(label_symbol(Eq(), 0), 1, label_symbol)
        loss = label_symbol(Mul(),input_symbol)(ReduceSum(axis=1))(Plus(),(dimension - 1))(Max(), 0)(ReduceMean())
        return loss


class MSELoss(LossLayer):
    loss_type = LossCategory.regression

    @staticmethod
    def loss_function(input_symbol: V, target_symbol: V):
        loss = input_symbol(Sub(),target_symbol)(Pow(),2)(ReduceMean())
        return loss


softmax_loss = SoftMaxLoss.loss_function
svm_loss = SVMLoss.loss_function
mse_loss = MSELoss.loss_function


def softmax_loss_with_label(input_symbol: V, classification):
    label_symbol = constant(generate_label_matrix(classification)[0])
    return softmax_loss(input_symbol, label_symbol)


def svm_loss_with_label(input_symbol: V, classification):
    label_symbol = constant(generate_label_matrix(classification)[0])
    return svm_loss(input_symbol, label_symbol)


loss_map = {
    'softmax': SoftMaxLoss,
    'svm': SVMLoss,
    'mse': MSELoss,
}


def register_loss(name: str, loss: LossLayer):
    loss_map[name.lower()] = loss


class Loss:
    def __init__(self, name: str, *args, **kwargs):
        self.__name = name.lower()
        self.__loss = None
        if self.__name in loss_map:
            self.__loss = loss_map[self.__name](*args, **kwargs)
        else:
            raise ValueError('No such loss: {}'.format(name))

    def loss_layer(self):
        return self.__loss



from ..sheng import V,Pow,Mul,ReduceSum,Abs


class RegularizationLayer:
    def regularization_term(self, *args, **kwargs):
        pass


class RegularizationL1(RegularizationLayer):
    @staticmethod
    def regularization_term(target_symbol: V, decay: float):
        return target_symbol(Abs())(ReduceSum())(Mul(),decay)


class RegularizationL2(RegularizationLayer):
    @staticmethod
    def regularization_term(target_symbol: V, decay: float):
        return target_symbol(Pow(),2)(ReduceSum())(Pow(),0.5)(Mul(),decay)


regularization_l1 = RegularizationL1.regularization_term
regularization_l2 = RegularizationL2.regularization_term


regularization_map = {
    'l1': RegularizationL1,
    'l2': RegularizationL2,
}


def register_regularization(name: str, regularization: RegularizationLayer):
    regularization_map[name.lower()] = regularization


class Regularization:
    def __init__(self, name: str, *args, **kwargs):
        self.__name = name.lower()
        self.__regularization = None
        if self.__name in regularization_map:
            self.__regularization = regularization_map[self.__name](*args, **kwargs)
        else:
            raise ValueError('No such regularization: {}'.format(name))

    def regularization_layer(self):
        return self.__regularization



# -*- coding:utf-8 -*-
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

long_description ='''
##mini深度框架
https://hellotrik.github.io/trik
'''

install_requires = [
    'numpy>=1.13',
]

if __name__ == '__main__':
    setup(name='trik',
          version="0.0.4",
          author='hellotrik',
          url='https://space.bilibili.com/489142974',
          author_email="hellotrik@foxmail.com",
          description='改自paradox的 练习用 深度框架',
          long_description_content_type="text/markdown",
          long_description=long_description,          
          platforms=['MS Windows', 'Mac X', 'Unix/Linux'],
          license='MIT',
          keywords=['深度学习', '机器学习', '自动微分', '符号计算'],
          packages=['trik',
                    'trik.sheng',
                    'trik.ya',
                    'trik.sheng.conv',
                    'trik.yu',
                    'trik.yu.data_handler',
                    'trik.sheng.lstm',
                    'trik.yu.data_set'],
          install_requires=install_requires,
          classifiers=["Natural Language :: English",
                       "Programming Language :: Python",
                       "Operating System :: Microsoft :: Windows",
                       "Operating System :: Unix",
                       "Operating System :: MacOS",
                       "Programming Language :: Python :: 3",
                       ],
          zip_safe=True)
